/// <reference types="node" />
import { PrivateKey, ProtocolAddress, PublicKey, SenderCertificate, SenderKeyRecord, SenderKeyStore as SenderKeyStoreBase, SessionRecord, SessionStore as SessionStoreBase, SignedPreKeyRecord, Uuid } from '@signalapp/libsignal-client';
import { ProfileKey, ProfileKeyCredentialRequest, ServerPublicParams } from '@signalapp/libsignal-client/zkgroup';
import { signalservice as Proto } from '../../protos/compiled';
import { DeviceId, UUID, UUIDKind } from '../types';
import { Contact } from '../data/contacts';
import { Group as GroupData } from '../data/group';
import { EnvelopeType, ModifyGroupOptions, ModifyGroupResult, StorageWriteResult } from '../server/base';
import { ServerGroup } from '../server/group';
import { ChangeNumberOptions, Device, DeviceKeys, SingleUseKey } from '../data/device';
import { Group } from './group';
import { StorageState } from './storage-state';
export declare type Config = Readonly<{
    profileName: string;
    contacts: Proto.IAttachmentPointer;
    groups: Proto.IAttachmentPointer;
    trustRoot: PublicKey;
    serverPublicParams: ServerPublicParams;
    generateNumber(): Promise<string>;
    generateUUID(): Promise<UUID>;
    releaseUUID(uuid: UUID): Promise<void>;
    changeDeviceNumber(device: Device, options: ChangeNumberOptions): Promise<void>;
    send(device: Device, message: Buffer): Promise<void>;
    getSenderCertificate(): Promise<SenderCertificate>;
    getDeviceByUUID(uuid: UUID, deviceId?: DeviceId): Promise<Device | undefined>;
    issueExpiringProfileKeyCredential(device: Device, request: ProfileKeyCredentialRequest): Promise<Buffer | undefined>;
    getGroup(publicParams: Buffer): Promise<ServerGroup | undefined>;
    createGroup(group: Proto.IGroup): Promise<ServerGroup>;
    modifyGroup(options: ModifyGroupOptions): Promise<ModifyGroupResult>;
    waitForGroupUpdate(group: GroupData): Promise<void>;
    getStorageManifest(): Promise<Proto.IStorageManifest | undefined>;
    getStorageItem(key: Buffer): Promise<Buffer | undefined>;
    waitForStorageManifest(afterVersion?: number): Promise<void>;
    applyStorageWrite(operation: Proto.IWriteOperation, shouldNotify?: boolean): Promise<StorageWriteResult>;
}>;
export declare type EncryptOptions = Readonly<{
    timestamp?: number;
    sealed?: boolean;
    uuidKind?: UUIDKind;
    updatedPni?: UUID;
}>;
export declare type EncryptTextOptions = EncryptOptions & Readonly<{
    group?: Group;
    withProfileKey?: boolean;
    withPniSignature?: boolean;
}>;
export declare type CreateGroupOptions = Readonly<{
    title: string;
    members: ReadonlyArray<PrimaryDevice>;
}>;
export declare type InviteToGroupOptions = EncryptOptions & Readonly<{
    sendInvite?: boolean;
}>;
export declare type SyncSentOptions = Readonly<{
    timestamp: number;
    destinationUUID: UUID;
}>;
export declare type FetchStorageOptions = Readonly<{
    timestamp: number;
}>;
export declare type SendStickerPackSyncOptions = Readonly<{
    type: 'install' | 'remove';
    packId: Buffer;
    packKey: Buffer;
    timestamp?: number;
}>;
export declare type SyncReadMessage = Readonly<{
    senderUUID: UUID;
    timestamp: number;
}>;
export declare type SyncReadOptions = Readonly<{
    timestamp?: number;
    messages: ReadonlyArray<SyncReadMessage>;
}>;
export declare enum ReceiptType {
    Delivery = "Delivery",
    Read = "Read"
}
export declare type ReceiptOptions = Readonly<{
    timestamp?: number;
    type: ReceiptType;
    messageTimestamps: ReadonlyArray<number>;
}>;
export declare type UnencryptedReceiptOptions = Readonly<{
    timestamp?: number;
    messageTimestamp: number;
}>;
export declare type MessageQueueEntry = Readonly<{
    source: Device;
    uuidKind: UUIDKind;
    envelopeType: EnvelopeType;
    body: string;
    dataMessage: Proto.IDataMessage;
    content: Proto.IContent;
}>;
export declare type StoryQueueEntry = Readonly<{
    source: Device;
    uuidKind: UUIDKind;
    envelopeType: EnvelopeType;
    storyMessage: Proto.IStoryMessage;
    content: Proto.IContent;
}>;
export declare type SyncMessageQueueEntry = Readonly<{
    source: Device;
    syncMessage: Proto.ISyncMessage;
}>;
export declare type PrepareChangeNumberEntry = Readonly<{
    device: Device;
    envelope: Buffer;
}>;
export declare type PrepareChangeNumberResult = ReadonlyArray<PrepareChangeNumberEntry>;
export declare type ToContactOptions = Readonly<{
    includeProfileKey?: boolean;
}>;
export declare class SessionStore extends SessionStoreBase {
    private readonly sessions;
    saveSession(name: ProtocolAddress, record: SessionRecord): Promise<void>;
    getSession(name: ProtocolAddress): Promise<SessionRecord | null>;
    getExistingSessions(addresses: ProtocolAddress[]): Promise<SessionRecord[]>;
}
export declare class SenderKeyStore extends SenderKeyStoreBase {
    private readonly keys;
    saveSenderKey(sender: ProtocolAddress, distributionId: Uuid, record: SenderKeyRecord): Promise<void>;
    getSenderKey(sender: ProtocolAddress, distributionId: Uuid): Promise<SenderKeyRecord | null>;
}
export declare class PrimaryDevice {
    readonly device: Device;
    private readonly config;
    private isInitialized;
    private lockPromise;
    private readonly syncStates;
    private readonly storageKey;
    private readonly privateKey;
    private pniPrivateKey;
    private readonly contactsBlob;
    private readonly groupsBlob;
    private privSenderCertificate;
    private readonly messageQueue;
    private readonly storyQueue;
    private readonly syncMessageQueue;
    private privPniPublicKey;
    private readonly signedPreKeys;
    private readonly preKeys;
    private readonly sessions;
    private readonly senderKeys;
    private readonly identity;
    readonly signedPreKeyId: number;
    readonly publicKey: PublicKey;
    readonly profileKey: ProfileKey;
    readonly profileName: string;
    readonly secondaryDevices: Device[];
    readonly userAgent = "OWI";
    constructor(device: Device, config: Config);
    init(preKeyCount?: number): Promise<void>;
    toContact({ includeProfileKey, }?: ToContactOptions): Contact;
    addSecondaryDevice(device: Device): void;
    generateKeys(device: Device, uuidKind: UUIDKind, preKeyCount?: number): Promise<DeviceKeys & {
        signedPreKeyRecord: SignedPreKeyRecord;
    }>;
    getIdentityKey(uuidKind: UUIDKind): Promise<PrivateKey>;
    getPublicKey(uuidKind: UUIDKind): PublicKey;
    addSingleUseKey(target: Device, key: SingleUseKey, uuidKind?: UUIDKind): Promise<void>;
    getAllGroups(storage: StorageState): Promise<ReadonlyArray<Group>>;
    createGroup({ title, members: memberDevices }: CreateGroupOptions): Promise<Group>;
    waitForGroupUpdate(group: Group): Promise<Group>;
    inviteToGroup(group: Group, device: Device, options?: InviteToGroupOptions): Promise<Group>;
    acceptPniInvite(group: Group, device: Device, options?: EncryptOptions): Promise<Group>;
    waitForStorageState({ after }?: {
        after?: StorageState;
    }): Promise<StorageState>;
    getStorageState(): Promise<StorageState | undefined>;
    expectStorageState(reason: string): Promise<StorageState>;
    setStorageState(state: StorageState): Promise<StorageState>;
    waitForSync(secondaryDevice: Device): Promise<void>;
    resetSyncState(secondaryDevice: Device): void;
    handleEnvelope(source: Device | undefined, uuidKind: UUIDKind, envelopeType: EnvelopeType, encrypted: Buffer): Promise<void>;
    encryptText(target: Device, text: string, options?: EncryptTextOptions): Promise<Buffer>;
    encryptSyncSent(target: Device, text: string, options: SyncSentOptions): Promise<Buffer>;
    encryptSyncRead(target: Device, options: SyncReadOptions): Promise<Buffer>;
    sendFetchStorage(options: FetchStorageOptions): Promise<void>;
    sendStickerPackSync(options: SendStickerPackSyncOptions): Promise<void>;
    encryptReceipt(target: Device, options: ReceiptOptions): Promise<Buffer>;
    sendReceipt(target: Device, options: ReceiptOptions): Promise<void>;
    sendUnencryptedReceipt(target: Device, { messageTimestamp, timestamp }: UnencryptedReceiptOptions): Promise<void>;
    prepareChangeNumber(options?: EncryptOptions): Promise<PrepareChangeNumberResult>;
    sendChangeNumber(result: PrepareChangeNumberResult): Promise<void>;
    changeNumber(options?: EncryptOptions): Promise<void>;
    sendText(target: Device, text: string, options?: EncryptTextOptions): Promise<void>;
    receive(source: Device, encrypted: Buffer): Promise<void>;
    waitForMessage(): Promise<MessageQueueEntry>;
    waitForStory(): Promise<StoryQueueEntry>;
    waitForSyncMessage(predicate?: ((entry: SyncMessageQueueEntry) => boolean)): Promise<SyncMessageQueueEntry>;
    private getProfileKeyPresentation;
    private getPrivateKey;
    private encryptContent;
    private broadcast;
    private getSyncState;
    private handleSync;
    private handleDataMessage;
    private handleStoryMessage;
    private encrypt;
    private decrypt;
    private lock;
    private get senderCertificate();
    private processSenderKeyDistribution;
    private convertManifestToStorageState;
}
