"use strict";
// Copyright 2022 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PrimaryDevice = exports.SenderKeyStore = exports.SessionStore = exports.ReceiptType = void 0;
const assert_1 = __importDefault(require("assert"));
const crypto_1 = __importDefault(require("crypto"));
const long_1 = __importDefault(require("long"));
const libsignal_client_1 = require("@signalapp/libsignal-client");
const SignalClient = __importStar(require("@signalapp/libsignal-client"));
const debug_1 = __importDefault(require("debug"));
const zkgroup_1 = require("@signalapp/libsignal-client/zkgroup");
const uuid_1 = require("uuid");
const compiled_1 = require("../../protos/compiled");
const constants_1 = require("../constants");
const types_1 = require("../types");
const crypto_2 = require("../crypto");
const base_1 = require("../server/base");
const util_1 = require("../util");
const group_1 = require("./group");
const storage_state_1 = require("./storage-state");
const debug = (0, debug_1.default)('mock:primary-device');
var ReceiptType;
(function (ReceiptType) {
    ReceiptType["Delivery"] = "Delivery";
    ReceiptType["Read"] = "Read";
})(ReceiptType = exports.ReceiptType || (exports.ReceiptType = {}));
var SyncState;
(function (SyncState) {
    SyncState[SyncState["Empty"] = 0] = "Empty";
    SyncState[SyncState["Contacts"] = 1] = "Contacts";
    SyncState[SyncState["Groups"] = 2] = "Groups";
    SyncState[SyncState["Blocked"] = 4] = "Blocked";
    SyncState[SyncState["Configuration"] = 8] = "Configuration";
    SyncState[SyncState["Keys"] = 16] = "Keys";
    SyncState[SyncState["Complete"] = 31] = "Complete";
})(SyncState || (SyncState = {}));
class SignedPreKeyStore extends libsignal_client_1.SignedPreKeyStore {
    constructor() {
        super(...arguments);
        this.records = new Map();
    }
    async saveSignedPreKey(id, record) {
        this.records.set(id, record);
    }
    async getSignedPreKey(id) {
        const result = this.records.get(id);
        if (!result) {
            throw new Error(`Signed pre key not found: ${id}`);
        }
        return result;
    }
}
class PreKeyStore extends libsignal_client_1.PreKeyStore {
    constructor() {
        super(...arguments);
        this.records = new Map();
    }
    async savePreKey(id, record) {
        this.records.set(id, record);
    }
    async getPreKey(id) {
        const record = this.records.get(id);
        if (!record) {
            throw new Error(`Pre key not found: ${id}`);
        }
        return record;
    }
    async removePreKey(id) {
        this.records.delete(id);
    }
}
class IdentityStore extends libsignal_client_1.IdentityKeyStore {
    constructor(privateKey, registrationId) {
        super();
        this.privateKey = privateKey;
        this.registrationId = registrationId;
        this.knownIdentities = new Map();
    }
    async getIdentityKey() {
        return this.privateKey;
    }
    async getLocalRegistrationId() {
        return this.registrationId;
    }
    async saveIdentity(name, key) {
        this.knownIdentities.set((0, util_1.addressToString)(name), key);
        return true;
    }
    async isTrustedIdentity() {
        // We trust everyone
        return true;
    }
    async getIdentity(name) {
        return this.knownIdentities.get((0, util_1.addressToString)(name)) ||
            null;
    }
    // Not part of IdentityKeyStore API
    async updateIdentityKey(privateKey) {
        this.privateKey = privateKey;
    }
    async updateLocalRegistrationId(registrationId) {
        this.registrationId = registrationId;
    }
}
class SessionStore extends libsignal_client_1.SessionStore {
    constructor() {
        super(...arguments);
        this.sessions = new Map();
    }
    async saveSession(name, record) {
        this.sessions.set((0, util_1.addressToString)(name), record);
    }
    async getSession(name) {
        return this.sessions.get((0, util_1.addressToString)(name)) || null;
    }
    async getExistingSessions(addresses) {
        return addresses.map((name) => {
            const existing = this.sessions.get((0, util_1.addressToString)(name));
            if (!existing) {
                throw new Error('Existing session not found');
            }
            return existing;
        });
    }
}
exports.SessionStore = SessionStore;
class SenderKeyStore extends libsignal_client_1.SenderKeyStore {
    constructor() {
        super(...arguments);
        this.keys = new Map();
    }
    async saveSenderKey(sender, distributionId, record) {
        this.keys.set(`${(0, util_1.addressToString)(sender)}.${distributionId}`, record);
    }
    async getSenderKey(sender, distributionId) {
        const key = this.keys.get(`${(0, util_1.addressToString)(sender)}.${distributionId}`);
        return key || null;
    }
}
exports.SenderKeyStore = SenderKeyStore;
class PrimaryDevice {
    constructor(device, config) {
        this.device = device;
        this.config = config;
        this.isInitialized = false;
        this.syncStates = new WeakMap();
        this.storageKey = crypto_1.default.randomBytes(16);
        this.privateKey = libsignal_client_1.PrivateKey.generate();
        this.pniPrivateKey = libsignal_client_1.PrivateKey.generate();
        this.messageQueue = new util_1.PromiseQueue();
        this.storyQueue = new util_1.PromiseQueue();
        this.syncMessageQueue = new util_1.PromiseQueue();
        this.privPniPublicKey = this.pniPrivateKey.getPublicKey();
        // Various stores
        this.signedPreKeys = new Map();
        this.preKeys = new Map();
        this.sessions = new SessionStore();
        this.senderKeys = new Map();
        this.identity = new Map();
        this.signedPreKeyId = 1;
        this.publicKey = this.privateKey.getPublicKey();
        this.secondaryDevices = new Array();
        // TODO(indutny): make primary device type configurable
        this.userAgent = 'OWI';
        for (const uuidKind of [types_1.UUIDKind.ACI, types_1.UUIDKind.PNI]) {
            this.identity.set(uuidKind, new IdentityStore(uuidKind === types_1.UUIDKind.ACI ? this.privateKey : this.pniPrivateKey, this.device.getRegistrationId(uuidKind)));
            this.preKeys.set(uuidKind, new PreKeyStore());
            this.signedPreKeys.set(uuidKind, new SignedPreKeyStore());
            this.senderKeys.set(uuidKind, new SenderKeyStore());
        }
        this.contactsBlob = this.config.contacts;
        this.groupsBlob = this.config.groups;
        this.profileName = config.profileName;
        this.profileKey = new zkgroup_1.ProfileKey(crypto_1.default.randomBytes(32));
        this.device.profileName = (0, crypto_2.encryptProfileName)(this.profileKey.serialize(), this.profileName);
    }
    async init(preKeyCount) {
        if (this.isInitialized) {
            throw new Error('Already initialized');
        }
        for (const uuidKind of [types_1.UUIDKind.ACI, types_1.UUIDKind.PNI]) {
            const identity = this.identity.get(uuidKind);
            assert_1.default.ok(identity);
            await identity.saveIdentity(this.device.getAddressByKind(uuidKind), this.getPublicKey(uuidKind));
            await this.device.setKeys(uuidKind, await this.generateKeys(this.device, uuidKind, preKeyCount));
        }
        this.privSenderCertificate = await this.config.getSenderCertificate();
        this.device.profileKeyCommitment = this.profileKey.getCommitment(this.device.uuid);
        this.device.accessKey = (0, crypto_2.deriveAccessKey)(this.profileKey.serialize());
        this.isInitialized = true;
    }
    toContact({ includeProfileKey = true, } = {}) {
        return {
            uuid: this.device.uuid,
            number: this.device.number,
            profileName: this.profileName,
            profileKey: includeProfileKey ? this.profileKey.serialize() : undefined,
        };
    }
    addSecondaryDevice(device) {
        this.secondaryDevices.push(device);
        device.profileName = this.device.profileName;
        device.profileKeyCommitment = this.device.profileKeyCommitment;
        device.accessKey = this.device.accessKey;
    }
    //
    // Keys
    //
    async generateKeys(device, uuidKind, preKeyCount = constants_1.INITIAL_PREKEY_COUNT) {
        const signedPreKey = libsignal_client_1.PrivateKey.generate();
        const signedPreKeySig = this.getPrivateKey(uuidKind).sign(signedPreKey.getPublicKey().serialize());
        const shouldSave = device === this.device;
        const signedPreKeyRecord = libsignal_client_1.SignedPreKeyRecord.new(this.signedPreKeyId, Date.now(), signedPreKey.getPublicKey(), signedPreKey, signedPreKeySig);
        if (shouldSave) {
            await this.signedPreKeys.get(uuidKind)?.saveSignedPreKey(this.signedPreKeyId, signedPreKeyRecord);
        }
        // NOTE: it is important to start with `1` here
        const preKeys = [];
        for (let i = 1; i <= preKeyCount; i++) {
            const preKey = libsignal_client_1.PrivateKey.generate();
            const publicKey = preKey.getPublicKey();
            const record = libsignal_client_1.PreKeyRecord.new(i, publicKey, preKey);
            if (shouldSave) {
                await this.preKeys.get(uuidKind)?.savePreKey(i, record);
            }
            preKeys.push({ keyId: i, publicKey });
        }
        return {
            identityKey: this.getPublicKey(uuidKind),
            signedPreKey: {
                keyId: this.signedPreKeyId,
                publicKey: signedPreKey.getPublicKey(),
                signature: signedPreKeySig,
            },
            preKeys,
            signedPreKeyRecord: signedPreKeyRecord,
        };
    }
    async getIdentityKey(uuidKind) {
        const identity = this.identity.get(uuidKind);
        assert_1.default.ok(identity);
        return identity.getIdentityKey();
    }
    getPublicKey(uuidKind) {
        switch (uuidKind) {
            case types_1.UUIDKind.ACI:
                return this.publicKey;
            case types_1.UUIDKind.PNI:
                return this.privPniPublicKey;
        }
    }
    async addSingleUseKey(target, key, uuidKind = types_1.UUIDKind.ACI) {
        assert_1.default.ok(this.isInitialized, 'Not initialized');
        debug('adding singleUseKey for', target.debugId);
        // Outgoing stores
        const identity = this.identity.get(types_1.UUIDKind.ACI);
        (0, assert_1.default)(identity, 'Should have an ACI identity');
        await identity.saveIdentity(target.getAddressByKind(uuidKind), key.identityKey);
        const bundle = libsignal_client_1.PreKeyBundle.new(target.getRegistrationId(uuidKind), target.deviceId, key.preKey === undefined ? null : key.preKey.keyId, key.preKey === undefined ? null : key.preKey.publicKey, key.signedPreKey.keyId, key.signedPreKey.publicKey, key.signedPreKey.signature, key.identityKey);
        await SignalClient.processPreKeyBundle(bundle, target.getAddressByKind(uuidKind), this.sessions, identity);
    }
    //
    // Groups
    //
    async getAllGroups(storage) {
        const records = storage.getAllGroupRecords();
        return await Promise.all(records.map(async ({ record }) => {
            const { groupV2 } = record;
            assert_1.default.ok(groupV2, 'Not a group v2 record!');
            const { masterKey } = groupV2;
            assert_1.default.ok(masterKey, 'Group v2 record without master key');
            const secretParams = zkgroup_1.GroupSecretParams.deriveFromMasterKey(new zkgroup_1.GroupMasterKey(Buffer.from(masterKey)));
            const publicParams = secretParams.getPublicParams().serialize();
            const serverGroup = await this.config.getGroup(publicParams);
            assert_1.default.ok(serverGroup, `Group not found: ${publicParams.toString('base64')}`);
            return new group_1.Group({
                secretParams,
                groupState: serverGroup.state,
            });
        }));
    }
    async createGroup({ title, members: memberDevices }) {
        const groupParams = zkgroup_1.GroupSecretParams.generate();
        const members = await Promise.all(memberDevices.map(async (member) => {
            const presentation = await member.getProfileKeyPresentation(groupParams);
            return {
                uuid: member.device.uuid,
                profileKey: member.profileKey,
                presentation,
                joinedAtVersion: long_1.default.fromNumber(0),
            };
        }));
        const clientGroup = group_1.Group.fromConfig({
            secretParams: groupParams,
            title,
            members,
        });
        await this.config.createGroup(clientGroup.state);
        return clientGroup;
    }
    async waitForGroupUpdate(group) {
        await this.config.waitForGroupUpdate(group);
        const publicParams = group.publicParams.serialize();
        const serverGroup = await this.config.getGroup(publicParams);
        assert_1.default.ok(serverGroup, `Group not found: ${group.id}`);
        return new group_1.Group({
            secretParams: group.secretParams,
            groupState: serverGroup.state,
        });
    }
    async inviteToGroup(group, device, options = {}) {
        const { uuidKind = types_1.UUIDKind.ACI, sendInvite = true } = options;
        const serverGroup = await this.config.getGroup(group.publicParams.serialize());
        (0, assert_1.default)(serverGroup !== undefined, 'Group does not exist on server');
        const targetUUID = device.getUUIDByKind(uuidKind);
        const userId = group.encryptUUID(targetUUID);
        const modifyResult = await this.config.modifyGroup({
            group: serverGroup,
            actions: {
                version: group.revision + 1,
                addPendingMembers: [{
                        added: {
                            member: { userId, role: compiled_1.signalservice.Member.Role.DEFAULT },
                        },
                    }],
            },
            aciCiphertext: group.encryptUUID(this.device.uuid),
            pniCiphertext: group.encryptUUID(this.device.pni),
        });
        (0, assert_1.default)(!modifyResult.conflict, 'Group update conflict!');
        const updatedGroup = new group_1.Group({
            secretParams: group.secretParams,
            groupState: serverGroup.state,
        });
        if (sendInvite) {
            // Send the invitation
            const encryptOptions = {
                timestamp: Date.now(),
                ...options,
            };
            const envelope = await this.encryptContent(device, {
                dataMessage: {
                    groupV2: {
                        ...updatedGroup.toContext(),
                        groupChange: compiled_1.signalservice.GroupChange.encode(modifyResult.signedChange).finish(),
                    },
                    timestamp: long_1.default.fromNumber(encryptOptions.timestamp),
                },
            }, encryptOptions);
            await this.config.send(device, envelope);
        }
        return updatedGroup;
    }
    async acceptPniInvite(group, device, options = {}) {
        const serverGroup = await this.config.getGroup(group.publicParams.serialize());
        (0, assert_1.default)(serverGroup !== undefined, 'Group does not exist on server');
        const aciCiphertext = group.encryptUUID(this.device.uuid);
        const pniCiphertext = group.encryptUUID(this.device.pni);
        const presentation = await this.getProfileKeyPresentation(group.secretParams);
        const modifyResult = await this.config.modifyGroup({
            group: serverGroup,
            actions: {
                version: group.revision + 1,
                promoteMembersPendingPniAciProfileKey: [{
                        presentation: presentation.serialize(),
                    }],
            },
            aciCiphertext,
            pniCiphertext,
        });
        (0, assert_1.default)(!modifyResult.conflict, 'Group update conflict!');
        const updatedGroup = new group_1.Group({
            secretParams: group.secretParams,
            groupState: serverGroup.state,
        });
        // Send the invitation
        const encryptOptions = {
            timestamp: Date.now(),
            ...options,
        };
        const envelope = await this.encryptContent(device, {
            dataMessage: {
                groupV2: {
                    ...updatedGroup.toContext(),
                    groupChange: compiled_1.signalservice.GroupChange.encode(modifyResult.signedChange).finish(),
                },
                timestamp: long_1.default.fromNumber(encryptOptions.timestamp),
            },
        }, encryptOptions);
        await this.config.send(device, envelope);
        return updatedGroup;
    }
    //
    // Storage Service
    //
    async waitForStorageState({ after } = {}) {
        debug('waiting for storage manifest for device=%s after version=%d', this.device.debugId, after?.version);
        await this.config.waitForStorageManifest(after?.version);
        const state = await this.getStorageState();
        (0, assert_1.default)(state, 'Missing storage state');
        debug('got storage manifest for device=%s version=%d', this.device.debugId, state.version);
        return state;
    }
    async getStorageState() {
        const manifest = await this.config.getStorageManifest();
        if (!manifest) {
            return undefined;
        }
        return this.convertManifestToStorageState(manifest);
    }
    async expectStorageState(reason) {
        const state = await this.getStorageState();
        if (!state) {
            throw new Error(`expectStorageState: no storage state, ${reason}`);
        }
        return state;
    }
    async setStorageState(state) {
        const writeOperation = state.createWriteOperation(this.storageKey);
        (0, assert_1.default)(writeOperation.manifest, 'write operation without manifest');
        const { updated, error } = await this.config.applyStorageWrite(writeOperation, false);
        if (!updated) {
            throw new Error(`setStorageState: failed to update, ${error}`);
        }
        return this.convertManifestToStorageState(writeOperation.manifest);
    }
    //
    // Sync
    //
    // TODO(indutny): timeout
    async waitForSync(secondaryDevice) {
        debug('waiting for sync with %s', secondaryDevice.debugId);
        const { onComplete } = this.getSyncState(secondaryDevice);
        await onComplete;
    }
    resetSyncState(secondaryDevice) {
        this.syncStates.delete(secondaryDevice);
    }
    //
    // Receive/Send
    //
    async handleEnvelope(source, uuidKind, envelopeType, encrypted) {
        const { unsealedSource, content, envelopeType: unsealedType } = await this.lock(async () => {
            return await this.decrypt(source, uuidKind, envelopeType, encrypted);
        });
        let handled = true;
        if (content.syncMessage) {
            assert_1.default.strictEqual(uuidKind, types_1.UUIDKind.ACI, 'Got sync message on PNI');
            await this.handleSync(unsealedSource, content.syncMessage);
        }
        else if (content.dataMessage) {
            await this.handleDataMessage(unsealedSource, uuidKind, unsealedType, content);
        }
        else if (content.storyMessage) {
            await this.handleStoryMessage(unsealedSource, uuidKind, unsealedType, content);
        }
        else {
            handled = false;
        }
        const { senderKeyDistributionMessage } = content;
        if (senderKeyDistributionMessage &&
            senderKeyDistributionMessage.length > 0) {
            handled = true;
            await this.processSenderKeyDistribution(unsealedSource, senderKeyDistributionMessage);
        }
        if (!handled) {
            debug('unsupported message', content);
        }
    }
    async encryptText(target, text, options = {}) {
        const encryptOptions = {
            timestamp: Date.now(),
            ...options,
        };
        let pniSignatureMessage;
        if (options.withPniSignature) {
            const pniPrivate = this.getPrivateKey(types_1.UUIDKind.PNI);
            const pniPublic = this.getPublicKey(types_1.UUIDKind.PNI);
            const aciPublic = this.getPublicKey(types_1.UUIDKind.ACI);
            const pniIdentity = new libsignal_client_1.IdentityKeyPair(pniPublic, pniPrivate);
            const signature = pniIdentity.signAlternateIdentity(aciPublic);
            pniSignatureMessage = {
                pni: new Uint8Array((0, uuid_1.parse)(this.device.pni)),
                signature,
            };
        }
        const content = {
            dataMessage: {
                groupV2: options.group?.toContext(),
                body: text,
                profileKey: options.withProfileKey ?
                    this.profileKey.serialize() :
                    undefined,
                timestamp: long_1.default.fromNumber(encryptOptions.timestamp),
            },
            pniSignatureMessage,
        };
        return await this.encryptContent(target, content, encryptOptions);
    }
    async encryptSyncSent(target, text, options) {
        const dataMessage = {
            body: text,
            timestamp: long_1.default.fromNumber(options.timestamp),
        };
        const content = {
            syncMessage: {
                sent: {
                    destinationUuid: options.destinationUUID,
                    timestamp: long_1.default.fromNumber(options.timestamp),
                    message: dataMessage,
                },
            },
        };
        return await this.encryptContent(target, content, options);
    }
    async encryptSyncRead(target, options) {
        const content = {
            syncMessage: {
                read: options.messages.map(({ senderUUID, timestamp }) => {
                    return {
                        senderUuid: senderUUID,
                        timestamp: long_1.default.fromNumber(timestamp),
                    };
                }),
            },
        };
        return await this.encryptContent(target, content, options);
    }
    async sendFetchStorage(options) {
        const content = {
            syncMessage: {
                fetchLatest: {
                    type: compiled_1.signalservice.SyncMessage.FetchLatest.Type.STORAGE_MANIFEST,
                },
            },
        };
        return this.broadcast('fetch storage', content, options);
    }
    async sendStickerPackSync(options) {
        const Type = compiled_1.signalservice.SyncMessage.StickerPackOperation.Type;
        const content = {
            syncMessage: {
                stickerPackOperation: [{
                        packId: options.packId,
                        packKey: options.packKey,
                        type: options.type === 'install' ? Type.INSTALL : Type.REMOVE,
                    }],
            },
        };
        return this.broadcast('sticker pack sync', content, options);
    }
    async encryptReceipt(target, options) {
        let type;
        if (options.type === ReceiptType.Delivery) {
            type = compiled_1.signalservice.ReceiptMessage.Type.DELIVERY;
        }
        else {
            assert_1.default.strictEqual(options.type, ReceiptType.Read);
            type = compiled_1.signalservice.ReceiptMessage.Type.READ;
        }
        const content = {
            receiptMessage: {
                type,
                timestamp: options.messageTimestamps.map((timestamp) => long_1.default.fromNumber(timestamp)),
            },
        };
        return await this.encryptContent(target, content, options);
    }
    async sendReceipt(target, options) {
        const receipt = await this.encryptReceipt(target, options);
        return this.config.send(target, receipt);
    }
    async sendUnencryptedReceipt(target, { messageTimestamp, timestamp = Date.now() }) {
        const envelope = {
            type: compiled_1.signalservice.Envelope.Type.RECEIPT,
            timestamp: long_1.default.fromNumber(messageTimestamp),
            serverTimestamp: long_1.default.fromNumber(timestamp),
            sourceUuid: this.device.uuid,
            sourceDevice: this.device.deviceId,
            destinationUuid: target.uuid,
        };
        return this.config.send(target, Buffer.from(compiled_1.signalservice.Envelope.encode(envelope).finish()));
    }
    async prepareChangeNumber(options = {}) {
        const { timestamp = Date.now() } = options;
        const newNumber = await this.config.generateNumber();
        const newPni = await this.config.generateUUID();
        const newPniRegistrationId = (0, util_1.generateRegistrationId)();
        const newPniIdentity = libsignal_client_1.IdentityKeyPair.generate();
        debug('sending change number to %d linked devices timestamp=%d newPni=%s', this.secondaryDevices.length, timestamp, newPni);
        this.pniPrivateKey = newPniIdentity.privateKey;
        this.privPniPublicKey = newPniIdentity.publicKey;
        const allDevices = [this.device, ...this.secondaryDevices];
        const oldPni = this.device.pni;
        // Update PNI
        await Promise.all(allDevices.map(async (device) => {
            await this.config.changeDeviceNumber(device, {
                pni: newPni,
                number: newNumber,
                pniRegistrationId: newPniRegistrationId,
            });
        }));
        const identity = this.identity.get(types_1.UUIDKind.PNI);
        (0, assert_1.default)(identity, 'Should have a PNI identity');
        await identity.updateIdentityKey(newPniIdentity.privateKey);
        await identity.updateLocalRegistrationId(newPniRegistrationId);
        await identity.saveIdentity(this.device.getAddressByKind(types_1.UUIDKind.PNI), this.getPublicKey(types_1.UUIDKind.PNI));
        await this.config.releaseUUID(oldPni);
        // Update all keys and prepare sync message
        const results = await Promise.all(allDevices.map(async (device) => {
            const isPrimary = device === this.device;
            const preKeyCount = isPrimary ? undefined : 0;
            const keys = await this.generateKeys(device, types_1.UUIDKind.PNI, preKeyCount);
            await device.setKeys(types_1.UUIDKind.PNI, keys);
            if (isPrimary) {
                return;
            }
            // Send sync message
            const { signedPreKeyRecord } = keys;
            const content = {
                syncMessage: {
                    pniChangeNumber: {
                        identityKeyPair: newPniIdentity.serialize(),
                        signedPreKey: signedPreKeyRecord.serialize(),
                        registrationId: newPniRegistrationId,
                    },
                },
            };
            const envelope = await this.encryptContent(device, content, {
                ...options,
                timestamp,
                updatedPni: this.device.pni,
            });
            return { device, envelope };
        }));
        return results.filter((entry) => {
            return entry !== undefined;
        });
    }
    async sendChangeNumber(result) {
        await Promise.all(result.map(({ device, envelope }) => {
            return this.config.send(device, envelope);
        }));
    }
    async changeNumber(options = {}) {
        const result = await this.prepareChangeNumber(options);
        await this.sendChangeNumber(result);
    }
    async sendText(target, text, options) {
        await this.config.send(target, await this.encryptText(target, text, options));
    }
    async receive(source, encrypted) {
        const envelope = compiled_1.signalservice.Envelope.decode(encrypted);
        if (source.uuid !== envelope.sourceUuid) {
            throw new Error(`Invalid envelope source. Expected: ${source.uuid}, ` +
                `Got: ${envelope.sourceUuid}`);
        }
        let envelopeType;
        if (envelope.type === compiled_1.signalservice.Envelope.Type.CIPHERTEXT) {
            envelopeType = base_1.EnvelopeType.CipherText;
        }
        else if (envelope.type === compiled_1.signalservice.Envelope.Type.PREKEY_BUNDLE) {
            envelopeType = base_1.EnvelopeType.PreKey;
        }
        else if (envelope.type === compiled_1.signalservice.Envelope.Type.UNIDENTIFIED_SENDER) {
            envelopeType = base_1.EnvelopeType.SealedSender;
        }
        else {
            throw new Error('Unsupported envelope type');
        }
        const uuidKind = envelope.destinationUuid ?
            this.device.getUUIDKind(envelope.destinationUuid) :
            types_1.UUIDKind.ACI;
        return await this.handleEnvelope(source, uuidKind, envelopeType, Buffer.from(envelope.content));
    }
    async waitForMessage() {
        return this.messageQueue.shift();
    }
    async waitForStory() {
        return this.storyQueue.shift();
    }
    async waitForSyncMessage(predicate = () => true) {
        for (;;) {
            const entry = await this.syncMessageQueue.shift();
            if (!predicate(entry)) {
                continue;
            }
            return entry;
        }
    }
    //
    // Private
    //
    async getProfileKeyPresentation(groupParams) {
        const ops = new zkgroup_1.ClientZkProfileOperations(this.config.serverPublicParams);
        const ctx = ops.createProfileKeyCredentialRequestContext(this.device.uuid, this.profileKey);
        const response = await this.config.issueExpiringProfileKeyCredential(this.device, ctx.getRequest());
        assert_1.default.ok(response, `Member device ${this.device.uuid} not initialized`);
        const credential = ops.receiveExpiringProfileKeyCredential(ctx, new zkgroup_1.ExpiringProfileKeyCredentialResponse(response));
        return ops.createExpiringProfileKeyCredentialPresentation(groupParams, credential);
    }
    getPrivateKey(uuidKind) {
        switch (uuidKind) {
            case types_1.UUIDKind.ACI:
                return this.privateKey;
            case types_1.UUIDKind.PNI:
                return this.pniPrivateKey;
        }
    }
    async encryptContent(target, content, options) {
        const encoded = Buffer.from(compiled_1.signalservice.Content.encode(content).finish());
        return await this.lock(async () => {
            return await this.encrypt(target, encoded, options);
        });
    }
    async broadcast(type, content, options) {
        debug('broadcasting %s to %d linked devices', type, this.secondaryDevices.length);
        await Promise.all(this.secondaryDevices.map(async (device) => {
            const envelope = await this.encryptContent(device, content, options);
            await this.config.send(device, envelope);
        }));
    }
    getSyncState(secondaryDevice) {
        const existing = this.syncStates.get(secondaryDevice);
        if (existing) {
            return existing;
        }
        let complete;
        const onComplete = new Promise((resolve) => {
            complete = resolve;
        });
        if (!complete) {
            throw new Error('Failed to obtain resolve callback');
        }
        const entry = {
            state: SyncState.Empty,
            onComplete,
            complete,
        };
        this.syncStates.set(secondaryDevice, entry);
        return entry;
    }
    async handleSync(source, sync) {
        const { request } = sync;
        if (!request) {
            debug('got generic sync message');
            this.syncMessageQueue.push({
                source,
                syncMessage: sync,
            });
            return;
        }
        let stateChange;
        let response;
        if (request.type === compiled_1.signalservice.SyncMessage.Request.Type.CONTACTS) {
            debug('got sync contacts request');
            response = {
                contacts: {
                    blob: this.contactsBlob,
                    complete: true,
                },
            };
            stateChange = SyncState.Contacts;
        }
        else if (request.type === compiled_1.signalservice.SyncMessage.Request.Type.GROUPS) {
            debug('got sync groups request');
            response = {
                groups: {
                    blob: this.groupsBlob,
                },
            };
            stateChange = SyncState.Groups;
        }
        else if (request.type === compiled_1.signalservice.SyncMessage.Request.Type.BLOCKED) {
            debug('got sync blocked request');
            response = {
                blocked: {},
            };
            stateChange = SyncState.Blocked;
        }
        else if (request.type === compiled_1.signalservice.SyncMessage.Request.Type.CONFIGURATION) {
            debug('got sync configuration request');
            response = {
                configuration: {
                    readReceipts: true,
                    unidentifiedDeliveryIndicators: false,
                    typingIndicators: false,
                    linkPreviews: false,
                },
            };
            stateChange = SyncState.Configuration;
        }
        else if (request.type === compiled_1.signalservice.SyncMessage.Request.Type.KEYS) {
            debug('got sync keys request');
            response = {
                keys: { storageService: this.storageKey },
            };
            stateChange = SyncState.Keys;
        }
        else {
            debug('Unsupported sync request', request);
            return;
        }
        const encrypted = await this.encryptContent(source, {
            syncMessage: response,
        });
        await this.config.send(source, encrypted);
        const syncEntry = this.getSyncState(source);
        syncEntry.state |= stateChange;
        if (syncEntry.state === SyncState.Complete) {
            debug('sync with %s complete', source.debugId);
            syncEntry.complete();
        }
    }
    async handleDataMessage(source, uuidKind, envelopeType, content) {
        const { dataMessage } = content;
        assert_1.default.ok(dataMessage, 'dataMessage must be present');
        const { body } = dataMessage;
        this.messageQueue.push({
            source,
            uuidKind,
            body: body ?? '',
            envelopeType,
            dataMessage,
            content,
        });
    }
    async handleStoryMessage(source, uuidKind, envelopeType, content) {
        const { storyMessage } = content;
        assert_1.default.ok(storyMessage, 'storyMessage must be present');
        this.storyQueue.push({
            source,
            uuidKind,
            envelopeType,
            storyMessage,
            content,
        });
    }
    async encrypt(target, message, { timestamp = Date.now(), sealed = false, uuidKind = types_1.UUIDKind.ACI, updatedPni, } = {}) {
        assert_1.default.ok(this.isInitialized, 'Not initialized');
        // "Pad"
        const paddedMessage = Buffer.concat([
            message,
            Buffer.from([0x80]),
        ]);
        let envelopeType;
        let content;
        // Outgoing stores
        const identity = this.identity.get(types_1.UUIDKind.ACI);
        (0, assert_1.default)(identity, 'Should have an ACI identity');
        if (sealed) {
            (0, assert_1.default)(uuidKind === types_1.UUIDKind.ACI, 'Can\'t send sealed sender to PNI');
            content = await SignalClient.sealedSenderEncryptMessage(paddedMessage, target.getAddressByKind(uuidKind), this.senderCertificate, this.sessions, identity);
            envelopeType = compiled_1.signalservice.Envelope.Type.UNIDENTIFIED_SENDER;
        }
        else {
            const ciphertext = await SignalClient.signalEncrypt(paddedMessage, target.getAddressByKind(uuidKind), this.sessions, identity);
            content = ciphertext.serialize();
            if (ciphertext.type() === libsignal_client_1.CiphertextMessageType.Whisper) {
                (0, assert_1.default)(uuidKind === types_1.UUIDKind.ACI, 'Can\'t send non-prekey messages to PNI');
                envelopeType = compiled_1.signalservice.Envelope.Type.CIPHERTEXT;
                debug('encrypting ciphertext envelope');
            }
            else {
                assert_1.default.strictEqual(ciphertext.type(), libsignal_client_1.CiphertextMessageType.PreKey);
                envelopeType = compiled_1.signalservice.Envelope.Type.PREKEY_BUNDLE;
                debug('encrypting prekeyBundle envelope');
            }
        }
        const envelope = Buffer.from(compiled_1.signalservice.Envelope.encode({
            type: envelopeType,
            sourceUuid: this.device.uuid,
            sourceDevice: this.device.deviceId,
            destinationUuid: target.getUUIDByKind(uuidKind),
            updatedPni,
            serverTimestamp: long_1.default.fromNumber(timestamp),
            timestamp: long_1.default.fromNumber(timestamp),
            content,
        }).finish());
        debug('encrypting envelope finish');
        return envelope;
    }
    async decrypt(source, uuidKind, envelopeType, encrypted) {
        debug('decrypting envelope type=%s start', envelopeType);
        const identity = this.identity.get(uuidKind);
        const preKeys = this.preKeys.get(uuidKind);
        const signedPreKeys = this.signedPreKeys.get(uuidKind);
        const senderKeys = this.senderKeys.get(uuidKind);
        (0, assert_1.default)(identity && preKeys && signedPreKeys && senderKeys, 'Should have identity, prekey/signed prekey/senderkey stores');
        let decrypted;
        if (envelopeType === base_1.EnvelopeType.CipherText) {
            (0, assert_1.default)(source !== undefined, 'CipherText must have source');
            decrypted = await SignalClient.signalDecrypt(libsignal_client_1.SignalMessage.deserialize(encrypted), source.getAddressByKind(uuidKind), this.sessions, identity);
        }
        else if (envelopeType === base_1.EnvelopeType.PreKey) {
            (0, assert_1.default)(source !== undefined, 'PreKey must have source');
            decrypted = await SignalClient.signalDecryptPreKey(libsignal_client_1.PreKeySignalMessage.deserialize(encrypted), source.getAddressByKind(uuidKind), this.sessions, identity, preKeys, signedPreKeys);
        }
        else if (envelopeType === base_1.EnvelopeType.SenderKey) {
            (0, assert_1.default)(source !== undefined, 'SenderKey must have source');
            decrypted = await SignalClient.groupDecrypt(source.getAddressByKind(uuidKind), senderKeys, encrypted);
        }
        else if (envelopeType === base_1.EnvelopeType.SealedSender) {
            (0, assert_1.default)(source === undefined, 'Sealed sender must have no source');
            const usmc = await SignalClient.sealedSenderDecryptToUsmc(encrypted, identity);
            const unsealedType = usmc.msgType();
            const certificate = usmc.senderCertificate();
            const sender = await this.config.getDeviceByUUID(certificate.senderUuid(), certificate.senderDeviceId());
            (0, assert_1.default)(sender !== undefined, 'Unsealed sender not found');
            let subType;
            switch (unsealedType) {
                case libsignal_client_1.CiphertextMessageType.PreKey:
                    subType = base_1.EnvelopeType.PreKey;
                    break;
                case libsignal_client_1.CiphertextMessageType.Whisper:
                    subType = base_1.EnvelopeType.CipherText;
                    break;
                case libsignal_client_1.CiphertextMessageType.SenderKey:
                    subType = base_1.EnvelopeType.SenderKey;
                    break;
                default:
                    throw new Error(`Unsupported usmc type: ${unsealedType}`);
            }
            // TODO(indutny): use sealedSenderDecryptMessage once it will support
            // sender key.
            const result = await this.decrypt(sender, uuidKind, subType, usmc.contents());
            if (uuidKind === types_1.UUIDKind.PNI) {
                debug('sealed message on PNI', result);
                throw new Error('Got sealed message on PNI');
            }
            return result;
        }
        else {
            throw new Error(`Unsupported envelope type: ${envelopeType}`);
        }
        // Remove padding
        let padding = 1;
        while (decrypted[decrypted.length - padding] !== 0x80) {
            assert_1.default.strictEqual(decrypted[decrypted.length - padding], 0);
            padding++;
        }
        const content = compiled_1.signalservice.Content.decode(decrypted.slice(0, -padding));
        debug('decrypting envelope type=%s finish', envelopeType);
        return { unsealedSource: source, content, envelopeType };
    }
    async lock(callback) {
        while (this.lockPromise) {
            await this.lockPromise;
        }
        let unlock;
        this.lockPromise = new Promise((resolve) => {
            unlock = resolve;
        });
        try {
            return await callback();
        }
        finally {
            this.lockPromise = undefined;
            assert_1.default.ok(unlock);
            unlock();
        }
    }
    get senderCertificate() {
        if (!this.privSenderCertificate) {
            throw new Error('Sender certificate not set');
        }
        return this.privSenderCertificate;
    }
    async processSenderKeyDistribution(source, rawMessage) {
        const message = libsignal_client_1.SenderKeyDistributionMessage.deserialize(Buffer.from(rawMessage));
        const senderKeys = this.senderKeys.get(types_1.UUIDKind.ACI);
        (0, assert_1.default)(senderKeys, 'Should have a sender key store');
        debug('received SKDM from', source.debugId);
        await SignalClient.processSenderKeyDistributionMessage(source.address, message, senderKeys);
    }
    async convertManifestToStorageState(manifest) {
        const decryptedManifest = (0, crypto_2.decryptStorageManifest)(this.storageKey, manifest);
        (0, assert_1.default)(decryptedManifest.version, 'Consistency check');
        const version = decryptedManifest.version.toNumber();
        const items = await Promise.all((decryptedManifest.keys || []).map(async ({ type, raw: key }) => {
            (0, assert_1.default)(type !== null && type !== undefined, 'Missing manifestRecord.keys.type');
            (0, assert_1.default)(key, 'Missing manifestRecord.keys.raw');
            const keyBuffer = Buffer.from(key);
            const item = await this.config.getStorageItem(keyBuffer);
            if (!item) {
                throw new Error(`Missing item ${keyBuffer.toString('base64')}`);
            }
            return {
                type,
                key: keyBuffer,
                record: (0, crypto_2.decryptStorageItem)(this.storageKey, {
                    key,
                    value: item,
                }),
            };
        }));
        return new storage_state_1.StorageState(version, items);
    }
}
exports.PrimaryDevice = PrimaryDevice;
