"use strict";
// Copyright 2022 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Server = void 0;
const assert_1 = __importDefault(require("assert"));
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const https_1 = __importDefault(require("https"));
const url_1 = require("url");
const libsignal_client_1 = require("@signalapp/libsignal-client");
const zkgroup_1 = require("@signalapp/libsignal-client/zkgroup");
const debug_1 = __importDefault(require("debug"));
const ws_1 = __importDefault(require("ws"));
const micro_1 = require("micro");
const attachment_1 = require("../data/attachment");
const constants_1 = require("../constants");
const types_1 = require("../types");
const contacts_1 = require("../data/contacts");
const crypto_1 = require("../crypto");
const compiled_1 = require("../../protos/compiled");
const base_1 = require("../server/base");
const util_1 = require("../util");
const http_1 = require("../server/http");
const ws_2 = require("../server/ws");
const primary_device_1 = require("./primary-device");
const debug = (0, debug_1.default)('mock:server:mock');
const CERTS_DIR = path_1.default.join(__dirname, '..', '..', 'certs');
const CERT = fs_1.default.readFileSync(path_1.default.join(CERTS_DIR, 'full-cert.pem'));
const KEY = fs_1.default.readFileSync(path_1.default.join(CERTS_DIR, 'key.pem'));
const TRUST_ROOT = JSON.parse(fs_1.default.readFileSync(path_1.default.join(CERTS_DIR, 'trust-root.json')).toString());
const ZK_PARAMS = JSON.parse(fs_1.default.readFileSync(path_1.default.join(CERTS_DIR, 'zk-params.json')).toString());
const DEFAULT_API_TIMEOUT = 60000;
class Server extends base_1.Server {
    constructor(config = {}) {
        super();
        this.primaryDevices = new Map();
        this.knownNumbers = new Set();
        this.provisionResultQueueByCode = new Map();
        this.provisionResultQueueByKey = new Map();
        this.manifestQueueByUuid = new Map();
        this.groupQueueById = new Map();
        this.rateLimitCountByPair = new Map();
        this.unregisteredUuids = new Set();
        this.config = {
            timeout: DEFAULT_API_TIMEOUT,
            trustRoot: TRUST_ROOT,
            zkParams: ZK_PARAMS,
            ...config,
            https: {
                key: KEY,
                cert: CERT,
                ...(config.https || {}),
            },
        };
        const trustPrivate = Buffer.from(this.config.trustRoot.privateKey, 'base64');
        this.trustRoot = libsignal_client_1.PrivateKey.deserialize(trustPrivate);
        const zkSecret = Buffer.from(this.config.zkParams.secretParams, 'base64');
        this.zkSecret = new zkgroup_1.ServerSecretParams(zkSecret);
        this.certificate = (0, crypto_1.generateServerCertificate)(this.trustRoot);
        this.provisionQueue = this.createQueue();
    }
    async listen(port, host) {
        if (this.https) {
            throw new Error('Already listening');
        }
        const emptyData = (0, crypto_1.encryptAttachment)(Buffer.alloc(0));
        const emptyCDNKey = await this.storeAttachment(emptyData.blob);
        this.emptyAttachment = (0, attachment_1.attachmentToPointer)(emptyCDNKey, emptyData);
        const httpHandler = (0, http_1.createHandler)(this);
        const server = https_1.default.createServer(this.config.https || {}, (req, res) => {
            (0, micro_1.run)(req, res, httpHandler);
        });
        const wss = new ws_1.default.Server({ server });
        wss.on('connection', (ws, request) => {
            const conn = new ws_2.Connection(request, ws, this);
            conn.start().catch((error) => {
                ws.close();
                debug('Websocket handling error', error.stack);
            });
        });
        this.https = server;
        return await new Promise((resolve) => {
            server.listen(port, host, () => resolve());
        });
    }
    async close() {
        const https = this.https;
        if (!https) {
            throw new Error('Not listening');
        }
        debug('closing server');
        await new Promise((resolve) => https.close(resolve));
    }
    address() {
        if (!this.https) {
            throw new Error('Not listening');
        }
        const result = this.https.address();
        if (!result || typeof result !== 'object') {
            throw new Error('Invalid .address() result');
        }
        return result;
    }
    //
    // Various queues
    //
    async waitForProvision() {
        return await this.provisionQueue.shift();
    }
    async waitForStorageManifest(device, afterVersion) {
        let queue = this.manifestQueueByUuid.get(device.uuid);
        if (!queue) {
            queue = this.createQueue();
            this.manifestQueueByUuid.set(device.uuid, queue);
        }
        let version;
        do {
            version = await queue.shift();
        } while (afterVersion !== undefined && version <= afterVersion);
    }
    async waitForGroupUpdate(group) {
        let queue = this.groupQueueById.get(group.id);
        if (!queue) {
            queue = this.createQueue();
            this.groupQueueById.set(group.id, queue);
        }
        let version;
        do {
            version = await queue.shift();
        } while (version <= group.revision);
    }
    //
    // Helper methods
    //
    async createPrimaryDevice({ profileName, initialPreKeyCount, contacts = [], contactsWithoutProfileKey = [], }) {
        const number = await this.generateNumber();
        const uuid = await this.generateUUID();
        const pni = await this.generateUUID();
        const registrationId = await (0, util_1.generateRegistrationId)();
        const pniRegistrationId = await (0, util_1.generateRegistrationId)();
        const device = await this.registerDevice({
            uuid,
            pni,
            number,
            registrationId,
            pniRegistrationId,
        });
        debug('creating primary device with uuid=%s registrationId=%d', uuid, registrationId);
        if (!this.emptyAttachment) {
            throw new Error('Mock#init must be called before starting the server');
        }
        const contactsAttachment = (0, crypto_1.encryptAttachment)((0, contacts_1.serializeContacts)([
            ...contacts.map((device) => device.toContact()),
            ...contactsWithoutProfileKey.map((device) => device.toContact({
                includeProfileKey: false,
            })),
        ]));
        const contactsCDNKey = await this.storeAttachment(contactsAttachment.blob);
        debug('contacts cdn key', contactsCDNKey);
        debug('groups cdn key', this.emptyAttachment.cdnKey);
        const primary = new primary_device_1.PrimaryDevice(device, {
            profileName: profileName,
            contacts: (0, attachment_1.attachmentToPointer)(contactsCDNKey, contactsAttachment),
            groups: this.emptyAttachment,
            trustRoot: this.trustRoot.getPublicKey(),
            serverPublicParams: this.zkSecret.getPublicParams(),
            generateNumber: this.generateNumber.bind(this),
            generateUUID: this.generateUUID.bind(this),
            releaseUUID: this.releaseUUID.bind(this),
            changeDeviceNumber: this.changeDeviceNumber.bind(this),
            send: this.send.bind(this),
            getSenderCertificate: this.getSenderCertificate.bind(this, device),
            getDeviceByUUID: this.getDeviceByUUID.bind(this),
            issueExpiringProfileKeyCredential: this.issueExpiringProfileKeyCredential.bind(this),
            getGroup: this.getGroup.bind(this),
            createGroup: this.createGroup.bind(this),
            modifyGroup: this.modifyGroup.bind(this),
            waitForGroupUpdate: this.waitForGroupUpdate.bind(this),
            getStorageManifest: this.getStorageManifest.bind(this, device),
            getStorageItem: this.getStorageItem.bind(this, device),
            waitForStorageManifest: this.waitForStorageManifest.bind(this, device),
            applyStorageWrite: this.applyStorageWrite.bind(this, device),
        });
        await primary.init(initialPreKeyCount);
        this.primaryDevices.set(number, primary);
        this.primaryDevices.set(uuid, primary);
        debug('created primary device number=%s uuid=%s', number, uuid);
        return primary;
    }
    async createSecondaryDevice(primary) {
        const registrationId = await (0, util_1.generateRegistrationId)();
        const pniRegistrationId = await (0, util_1.generateRegistrationId)();
        const device = await this.registerDevice({
            uuid: primary.device.uuid,
            pni: primary.device.pni,
            number: primary.device.number,
            registrationId,
            pniRegistrationId,
        });
        for (const uuidKind of [types_1.UUIDKind.ACI, types_1.UUIDKind.PNI]) {
            await this.updateDeviceKeys(device, uuidKind, await primary.generateKeys(device, uuidKind));
        }
        primary.addSecondaryDevice(device);
        return device;
    }
    unregister(primary, uuidKind = types_1.UUIDKind.ACI) {
        this.unregisteredUuids.add(primary.device.getUUIDByKind(uuidKind));
    }
    register(primary, uuidKind = types_1.UUIDKind.ACI) {
        this.unregisteredUuids.delete(primary.device.getUUIDByKind(uuidKind));
    }
    rateLimit({ source, target }) {
        this.rateLimitCountByPair.set(`${source}:${target}`, 0);
    }
    stopRateLimiting({ source, target, }) {
        const key = `${source}:${target}`;
        const existing = this.rateLimitCountByPair.get(key);
        this.rateLimitCountByPair.delete(key);
        return existing;
    }
    //
    // Implement Server's abstract methods
    //
    async getProvisioningResponse(uuid) {
        const responseQueue = this.createQueue();
        const resultQueue = this.createQueue();
        await this.provisionQueue.pushAndWait({
            complete: async (response) => {
                await responseQueue.pushAndWait(response);
                return await resultQueue.shift();
            },
        });
        const { 
        // tsdevice:/?uuid=<uuid>&pub_key=<base64>
        provisionURL, primaryDevice, } = await responseQueue.shift();
        const query = (0, url_1.parse)(provisionURL, true).query || {};
        assert_1.default.strictEqual(query.uuid, uuid, 'UUID mismatch');
        if (!query.pub_key || Array.isArray(query.pub_key)) {
            throw new Error('Expected `pub_key` in provision URL');
        }
        const publicKey = libsignal_client_1.PublicKey.deserialize(Buffer.from(query.pub_key, 'base64'));
        const aciIdentityKey = await primaryDevice.getIdentityKey(types_1.UUIDKind.ACI);
        const pniIdentityKey = await primaryDevice.getIdentityKey(types_1.UUIDKind.PNI);
        const provisioningCode = await this.getProvisioningCode(uuid, primaryDevice.device.number);
        this.provisionResultQueueByCode.set(provisioningCode, {
            seenUuidKinds: new Set(),
            promiseQueue: resultQueue,
        });
        const envelopeData = compiled_1.signalservice.ProvisionMessage.encode({
            aciIdentityKeyPrivate: aciIdentityKey.serialize(),
            aciIdentityKeyPublic: aciIdentityKey.getPublicKey().serialize(),
            pniIdentityKeyPrivate: pniIdentityKey.serialize(),
            pniIdentityKeyPublic: pniIdentityKey.getPublicKey().serialize(),
            number: primaryDevice.device.number,
            aci: primaryDevice.device.uuid,
            pni: primaryDevice.device.pni,
            provisioningCode,
            profileKey: primaryDevice.profileKey.serialize(),
            userAgent: primaryDevice.userAgent,
            readReceipts: true,
            // TODO(indutny): is it correct?
            ProvisioningVersion: compiled_1.signalservice.ProvisioningVersion.CURRENT,
        }).finish();
        const { body, ephemeralKey } = (0, crypto_1.encryptProvisionMessage)(Buffer.from(envelopeData), publicKey);
        const envelope = compiled_1.signalservice.ProvisionEnvelope.encode({
            publicKey: ephemeralKey,
            body,
        }).finish();
        return { envelope: Buffer.from(envelope) };
    }
    async handleMessage(source, uuidKind, envelopeType, target, encrypted) {
        (0, assert_1.default)(source || envelopeType === base_1.EnvelopeType.SealedSender, 'No source for non-sealed sender envelope');
        debug('got message for %s.%d', target.uuid, target.deviceId);
        if (target.deviceId !== constants_1.PRIMARY_DEVICE_ID) {
            debug('ignoring message, not primary');
            return;
        }
        const primary = this.primaryDevices.get(target.uuid);
        if (!primary) {
            debug('ignoring message, primary device not found');
            return;
        }
        await primary.handleEnvelope(source, uuidKind, envelopeType, encrypted);
    }
    isUnregistered(uuid) {
        return this.unregisteredUuids.has(uuid);
    }
    isSendRateLimited({ source, target, }) {
        const key = `${source}:${target}`;
        const existing = this.rateLimitCountByPair.get(key);
        if (existing === undefined) {
            return false;
        }
        const newValue = existing + 1;
        debug('isSendRateLimited: source=%j target=%j count=%d', source, target, newValue);
        this.rateLimitCountByPair.set(key, newValue);
        return true;
    }
    //
    // Override `Server`'s methods to automatically pass keys to primary
    // devices.
    //
    // TODO(indutny): use popSingleUseKey() perhaps?
    //
    async updateDeviceKeys(device, uuidKind, keys) {
        await super.updateDeviceKeys(device, uuidKind, keys);
        const key = `${device.uuid}.${device.getRegistrationId(uuidKind)}`;
        // Device is marked as provisioned only once we have its keys
        const resultQueue = this.provisionResultQueueByKey.get(key);
        if (!resultQueue) {
            return;
        }
        debug('updateDeviceKeys: got keys for', device.debugId, uuidKind);
        const { seenUuidKinds, promiseQueue } = resultQueue;
        (0, assert_1.default)(!seenUuidKinds.has(uuidKind), `Duplicate uuid kind ${uuidKind} for device: ${device.debugId}`);
        seenUuidKinds.add(uuidKind);
        if (!seenUuidKinds.has(types_1.UUIDKind.ACI) ||
            !seenUuidKinds.has(types_1.UUIDKind.PNI)) {
            return;
        }
        this.provisionResultQueueByKey.delete(key);
        await promiseQueue.pushAndWait(device);
    }
    async provisionDevice(options) {
        const { provisioningCode } = options;
        const queue = this.provisionResultQueueByCode.get(provisioningCode);
        (0, assert_1.default)(queue !== undefined, `Missing provision result queue for code: ${provisioningCode}`);
        this.provisionResultQueueByCode.delete(provisioningCode);
        const device = await super.provisionDevice(options);
        for (const uuidKind of [types_1.UUIDKind.ACI, types_1.UUIDKind.PNI]) {
            const key = `${device.uuid}.${device.getRegistrationId(uuidKind)}`;
            this.provisionResultQueueByKey.set(key, queue);
        }
        const primary = this.primaryDevices.get(device.uuid);
        primary?.addSecondaryDevice(device);
        return device;
    }
    // Override `getStorageItems` to provide configurable limit for maximum
    // storage read keys.
    async getStorageItems(device, keys) {
        if (this.config.maxStorageReadKeys !== undefined &&
            keys.length > this.config.maxStorageReadKeys) {
            debug('getStorageItems: requested more than max keys', device.debugId);
            return undefined;
        }
        return super.getStorageItems(device, keys);
    }
    // Override updateGroup to notify about group modifications
    async modifyGroup(options) {
        const { group } = options;
        debug('modifyGroup', group.id);
        const result = await super.modifyGroup(options);
        let queue = this.groupQueueById.get(group.id);
        if (!queue) {
            queue = this.createQueue();
            this.groupQueueById.set(group.id, queue);
        }
        queue.push(group.revision);
        return result;
    }
    async onStorageManifestUpdate(device, version) {
        debug('onStorageManifestUpdate', device.debugId);
        let queue = this.manifestQueueByUuid.get(device.uuid);
        if (!queue) {
            queue = this.createQueue();
            this.manifestQueueByUuid.set(device.uuid, queue);
        }
        queue.push(version.toNumber());
    }
    //
    // Private
    //
    createQueue() {
        return new util_1.PromiseQueue({
            timeout: this.config.timeout,
        });
    }
    async generateNumber() {
        let number;
        do {
            number = (0, util_1.generateRandomE164)();
        } while (this.knownNumbers.has(number));
        this.knownNumbers.add(number);
        return number;
    }
}
exports.Server = Server;
