"use strict";
// Copyright 2022 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Device = void 0;
const debug_1 = __importDefault(require("debug"));
const libsignal_client_1 = require("@signalapp/libsignal-client");
const types_1 = require("../types");
const debug = (0, debug_1.default)('mock:device');
class Device {
    constructor(options) {
        this.keys = new Map();
        this.uuid = options.uuid;
        this.deviceId = options.deviceId;
        this.registrationId = options.registrationId;
        this.privPni = options.pni;
        this.privNumber = options.number;
        this.pniRegistrationId = options.pniRegistrationId;
        this.address = libsignal_client_1.ProtocolAddress.new(this.uuid, this.deviceId);
        this.privPniAddress = libsignal_client_1.ProtocolAddress.new(this.pni, this.deviceId);
    }
    get debugId() {
        return `${this.uuid}.${this.deviceId}`;
    }
    getRegistrationId(uuidKind) {
        switch (uuidKind) {
            case types_1.UUIDKind.ACI:
                return this.registrationId;
            case types_1.UUIDKind.PNI:
                return this.pniRegistrationId;
        }
    }
    get pni() {
        return this.privPni;
    }
    get number() {
        return this.privNumber;
    }
    get pniAddress() {
        return this.privPniAddress;
    }
    async changeNumber({ number, pni, pniRegistrationId, }) {
        this.privNumber = number;
        this.privPni = pni;
        this.pniRegistrationId = pniRegistrationId;
        this.privPniAddress = libsignal_client_1.ProtocolAddress.new(this.pni, this.deviceId);
    }
    async setKeys(uuidKind, keys) {
        debug('setting %s keys for %s', uuidKind, this.debugId);
        this.keys.set(uuidKind, {
            identityKey: keys.identityKey,
            signedPreKey: keys.signedPreKey,
            preKeys: keys.preKeys.slice(),
        });
    }
    async getIdentityKey(uuidKind = types_1.UUIDKind.ACI) {
        const keys = this.keys.get(uuidKind);
        if (!keys) {
            throw new Error('No keys available for device');
        }
        return keys.identityKey;
    }
    async popSingleUseKey(uuidKind = types_1.UUIDKind.ACI) {
        const keys = this.keys.get(uuidKind);
        if (!keys) {
            throw new Error('No keys available for device');
        }
        debug('popping single use key for %s', this.debugId);
        const preKey = keys.preKeys.shift();
        return {
            identityKey: keys.identityKey,
            signedPreKey: keys.signedPreKey,
            preKey,
        };
    }
    async getSingleUseKeyCount(uuidKind = types_1.UUIDKind.ACI) {
        const keys = this.keys.get(uuidKind);
        if (!keys) {
            throw new Error('No keys available for device');
        }
        return keys.preKeys.length;
    }
    getUUIDByKind(uuidKind) {
        switch (uuidKind) {
            case types_1.UUIDKind.ACI:
                return this.uuid;
            case types_1.UUIDKind.PNI:
                return this.pni;
        }
    }
    getUUIDKind(uuid) {
        if (uuid === this.uuid) {
            return types_1.UUIDKind.ACI;
        }
        if (uuid === this.pni) {
            return types_1.UUIDKind.PNI;
        }
        throw new Error(`Unknown uuid: ${uuid}`);
    }
    getAddressByKind(uuidKind) {
        switch (uuidKind) {
            case types_1.UUIDKind.ACI:
                return this.address;
            case types_1.UUIDKind.PNI:
                return this.pniAddress;
        }
    }
}
exports.Device = Device;
