/// <reference types="node" />
import Long from 'long';
import { SenderCertificate } from '@signalapp/libsignal-client';
import { AuthCredentialPresentation, ProfileKeyCredentialRequest, ServerSecretParams } from '@signalapp/libsignal-client/zkgroup';
import { signalservice as Proto } from '../../protos/compiled';
import { ServerCertificate } from '../crypto';
import { ChangeNumberOptions, Device, DeviceKeys } from '../data/device';
import { AttachmentId, DeviceId, ProvisioningCode, RegistrationId, UUID, UUIDKind } from '../types';
import { JSONMessage } from '../data/json.d';
import { ModifyGroupResult, ServerGroup } from './group';
export declare enum EnvelopeType {
    CipherText = "CipherText",
    PreKey = "PreKey",
    SealedSender = "SealedSender",
    SenderKey = "SenderKey"
}
export declare type ProvisioningResponse = Readonly<{
    envelope: Buffer;
}>;
export declare type GroupCredentialsRange = Readonly<{
    from: number;
    to: number;
}>;
export declare type StorageCredentials = Readonly<{
    username: string;
    password: string;
}>;
export declare type GroupCredentials = Array<{
    credential: string;
    redemptionTime: number;
}>;
export declare type PreparedMultiDeviceMessage = ReadonlyArray<[Device, JSONMessage]>;
export declare type ProvisionDeviceOptions = Readonly<{
    number: string;
    password: string;
    provisioningCode: ProvisioningCode;
    registrationId: RegistrationId;
    pniRegistrationId: RegistrationId;
}>;
export declare type RegisterDeviceOptions = Readonly<{
    uuid: UUID;
    pni: UUID;
    number: string;
    registrationId: RegistrationId;
    pniRegistrationId: RegistrationId;
}>;
export declare type PrepareMultiDeviceMessageResult = Readonly<{
    status: 'stale';
    staleDevices: ReadonlyArray<number>;
} | {
    status: 'incomplete';
    missingDevices: ReadonlyArray<number>;
    extraDevices: ReadonlyArray<number>;
} | {
    status: 'unknown';
} | {
    status: 'ok';
    targetUUID: UUID;
    result: PreparedMultiDeviceMessage;
}>;
export declare type StorageWriteResult = Readonly<{
    updated: false;
    manifest: Proto.IStorageManifest;
    error?: void;
} | {
    updated: true;
    manifest?: void;
    error?: void;
} | {
    updated?: void;
    error: string;
}>;
export declare type ModifyGroupOptions = Readonly<{
    group: ServerGroup;
    actions: Proto.GroupChange.IActions;
    aciCiphertext: Uint8Array;
    pniCiphertext: Uint8Array;
}>;
export declare type EncryptedStickerPack = Readonly<{
    id: Buffer;
    manifest: Buffer;
    stickers: ReadonlyArray<Buffer>;
}>;
export declare type IsSendRateLimitedOptions = Readonly<{
    source: UUID;
    target: UUID;
}>;
export { ModifyGroupResult };
interface WebSocket {
    sendMessage(message: Buffer | 'empty'): Promise<void>;
}
export declare abstract class Server {
    private readonly devices;
    private readonly devicesByUUID;
    private readonly usedUUIDs;
    private readonly devicesByAuth;
    private readonly storageAuthByUsername;
    private readonly storageAuthByDevice;
    private readonly storageManifestByUuid;
    private readonly storageItemsByUuid;
    private readonly provisioningCodes;
    private readonly attachments;
    private readonly stickerPacks;
    private readonly webSockets;
    private readonly messageQueue;
    private readonly groups;
    protected privCertificate: ServerCertificate | undefined;
    protected privZKSecret: ServerSecretParams | undefined;
    generateUUID(): Promise<UUID>;
    releaseUUID(uuid: UUID): Promise<void>;
    abstract getProvisioningResponse(uuid: UUID): Promise<ProvisioningResponse>;
    registerDevice({ uuid, pni, number, registrationId, pniRegistrationId, }: RegisterDeviceOptions): Promise<Device>;
    getProvisioningCode(uuid: UUID, number: string): Promise<ProvisioningCode>;
    provisionDevice({ number, password, provisioningCode, registrationId, pniRegistrationId, }: ProvisionDeviceOptions): Promise<Device>;
    updateDeviceKeys(device: Device, uuidKind: UUIDKind, keys: DeviceKeys): Promise<void>;
    changeDeviceNumber(device: Device, options: ChangeNumberOptions): Promise<void>;
    auth(username: string, password: string): Promise<Device | undefined>;
    protected storeAttachment(attachment: Buffer): Promise<AttachmentId>;
    fetchAttachment(id: AttachmentId): Promise<Buffer | undefined>;
    fetchStickerPack(packId: string): Promise<Buffer | undefined>;
    fetchSticker(packId: string, stickerId: number): Promise<Buffer | undefined>;
    storeStickerPack(pack: EncryptedStickerPack): Promise<void>;
    prepareMultiDeviceMessage(source: Device | undefined, targetUUID: UUID, messages: ReadonlyArray<JSONMessage>): Promise<PrepareMultiDeviceMessageResult>;
    handlePreparedMultiDeviceMessage(source: Device | undefined, targetUUID: UUID, prepared: PreparedMultiDeviceMessage): Promise<void>;
    abstract handleMessage(source: Device | undefined, uuidKind: UUIDKind, envelopeType: EnvelopeType, target: Device, encrypted: Buffer): Promise<void>;
    addWebSocket(device: Device, socket: WebSocket): Promise<void>;
    removeWebSocket(device: Device, socket: WebSocket): void;
    send(target: Device, message: Buffer): Promise<void>;
    createGroup(group: Proto.IGroup): Promise<ServerGroup>;
    modifyGroup({ group, actions, aciCiphertext, pniCiphertext, }: ModifyGroupOptions): Promise<ModifyGroupResult>;
    getGroup(publicParams: Buffer): Promise<ServerGroup | undefined>;
    getStorageAuth(device: Device): Promise<StorageCredentials>;
    storageAuth(username: string, password: string): Promise<Device | undefined>;
    getStorageManifest(device: Device): Promise<Proto.IStorageManifest | undefined>;
    applyStorageWrite(device: Device, { manifest, clearAll, insertItem, deleteKey, }: Proto.IWriteOperation, shouldNotify?: boolean): Promise<StorageWriteResult>;
    private clearStorageItems;
    private setStorageItem;
    getStorageItem(device: Device, key: Buffer): Promise<Buffer | undefined>;
    getStorageItems(device: Device, keys: ReadonlyArray<Buffer>): Promise<Array<Proto.IStorageItem> | undefined>;
    deleteStorageItem(device: Device, key: Buffer): Promise<void>;
    protected abstract onStorageManifestUpdate(device: Device, version: Long): Promise<void>;
    getDevice(number: string, deviceId: DeviceId): Promise<Device | undefined>;
    getDeviceByUUID(uuid: UUID, deviceId?: DeviceId): Promise<Device | undefined>;
    getAllDevicesByUUID(uuid: UUID): Promise<ReadonlyArray<Device>>;
    getSenderCertificate(device: Device): Promise<SenderCertificate>;
    getGroupCredentials({ uuid, pni }: Device, { from, to }: GroupCredentialsRange): Promise<GroupCredentials>;
    verifyGroupCredentials(publicParams: Buffer, credential: Buffer): Promise<AuthCredentialPresentation>;
    issueExpiringProfileKeyCredential({ uuid, profileKeyCommitment }: Device, request: ProfileKeyCredentialRequest): Promise<Buffer | undefined>;
    abstract isUnregistered(uuid: UUID): boolean;
    abstract isSendRateLimited(options: IsSendRateLimitedOptions): boolean;
    protected set certificate(value: ServerCertificate);
    protected get certificate(): ServerCertificate;
    protected set zkSecret(value: ServerSecretParams);
    protected get zkSecret(): ServerSecretParams;
    private sendQueue;
}
