"use strict";
// Copyright 2022 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Server = exports.EnvelopeType = void 0;
const assert_1 = __importDefault(require("assert"));
const crypto_1 = __importDefault(require("crypto"));
const uuid_1 = require("uuid");
const debug_1 = __importDefault(require("debug"));
const zkgroup_1 = require("@signalapp/libsignal-client/zkgroup");
const compiled_1 = require("../../protos/compiled");
const crypto_2 = require("../crypto");
const device_1 = require("../data/device");
const constants_1 = require("../constants");
const types_1 = require("../types");
const util_1 = require("../util");
const group_1 = require("./group");
var EnvelopeType;
(function (EnvelopeType) {
    EnvelopeType["CipherText"] = "CipherText";
    EnvelopeType["PreKey"] = "PreKey";
    EnvelopeType["SealedSender"] = "SealedSender";
    EnvelopeType["SenderKey"] = "SenderKey";
})(EnvelopeType = exports.EnvelopeType || (exports.EnvelopeType = {}));
const debug = (0, debug_1.default)('mock:server:base');
// NOTE: This class is currently extended only by src/api/server.ts
class Server {
    constructor() {
        this.devices = new Map();
        this.devicesByUUID = new Map();
        this.usedUUIDs = new Set();
        this.devicesByAuth = new Map();
        this.storageAuthByUsername = new Map();
        this.storageAuthByDevice = new Map();
        this.storageManifestByUuid = new Map();
        this.storageItemsByUuid = new Map();
        this.provisioningCodes = new Map();
        this.attachments = new Map();
        this.stickerPacks = new Map();
        this.webSockets = new Map();
        this.messageQueue = new WeakMap();
        this.groups = new Map();
    }
    //
    // Provisioning
    //
    async generateUUID() {
        let result;
        do {
            result = (0, uuid_1.v4)();
        } while (this.usedUUIDs.has(result) || this.devicesByUUID.has(result));
        this.usedUUIDs.add(result);
        return result;
    }
    async releaseUUID(uuid) {
        if (this.devicesByUUID.has(uuid)) {
            assert_1.default.ok(!this.usedUUIDs.has(uuid));
            throw new Error('Can\'t release UUID');
        }
        this.usedUUIDs.delete(uuid);
    }
    async registerDevice({ uuid, pni, number, registrationId, pniRegistrationId, }) {
        if (!this.usedUUIDs.has(uuid)) {
            throw new Error('Use generateUUID() to create new UUID');
        }
        let list = this.devices.get(number);
        if (!list) {
            list = [];
            this.devices.set(number, list);
        }
        const deviceId = list.length + 1;
        const isPrimary = deviceId === constants_1.PRIMARY_DEVICE_ID;
        const device = new device_1.Device({
            uuid,
            pni,
            number,
            deviceId,
            registrationId,
            pniRegistrationId,
        });
        if (isPrimary) {
            (0, assert_1.default)(!this.devicesByUUID.has(uuid), 'Duplicate primary device');
            this.devicesByUUID.set(uuid, device);
            this.devicesByUUID.set(pni, device);
        }
        list.push(device);
        debug('registered device number=%j uuid=%s', number, uuid);
        return device;
    }
    // Called from primary device
    async getProvisioningCode(uuid, number) {
        let entry = this.provisioningCodes.get(number);
        if (!entry) {
            entry = new Map();
            this.provisioningCodes.set(number, entry);
        }
        let code;
        do {
            code = crypto_1.default.randomBytes(8).toString('hex');
        } while (entry.has(code));
        entry.set(code, uuid);
        return code;
    }
    // Called from secondary device
    async provisionDevice({ number, password, provisioningCode, registrationId, pniRegistrationId, }) {
        const entry = this.provisioningCodes.get(number);
        if (!entry) {
            throw new Error('Invalid number for provisioning');
        }
        const uuid = entry.get(provisioningCode);
        if (!uuid) {
            throw new Error('Invalid provisioning code');
        }
        entry.delete(provisioningCode);
        const [primary] = this.devices.get(number) || [];
        (0, assert_1.default)(primary !== undefined, 'Missing primary device when provisioning');
        const device = await this.registerDevice({
            uuid: primary.uuid,
            pni: primary.pni,
            number,
            registrationId,
            pniRegistrationId,
        });
        const username = `${number}.${device.deviceId}`;
        // This is awkward, but WebSockets use it.
        const secondUsername = `${device.uuid}.${device.deviceId}`;
        // Add auth only after successfully registering the device
        (0, assert_1.default)(!this.devicesByAuth.has(username) &&
            !this.devicesByAuth.has(secondUsername), 'Duplicate username in `provisionDevice`');
        const authEntry = {
            password,
            device,
        };
        this.devicesByAuth.set(username, authEntry);
        this.devicesByAuth.set(secondUsername, authEntry);
        debug('provisioned device number=%j uuid=%j', number, uuid);
        return device;
    }
    async updateDeviceKeys(device, uuidKind, keys) {
        debug('setting device=%s keys', device.debugId);
        await device.setKeys(uuidKind, keys);
    }
    async changeDeviceNumber(device, options) {
        const oldNumber = device.number;
        const oldPni = device.pni;
        await device.changeNumber(options);
        const oldDevices = this.devices.get(oldNumber) ?? [];
        const oldDeviceIndex = oldDevices.indexOf(device);
        if (oldDeviceIndex !== -1) {
            oldDevices.splice(oldDeviceIndex, 1);
            if (oldDevices.length === 0) {
                this.devices.delete(oldNumber);
            }
        }
        let newDevices = this.devices.get(options.number);
        if (!newDevices) {
            newDevices = [];
            this.devices.set(options.number, newDevices);
        }
        newDevices.push(device);
        const oldPrimary = this.devicesByUUID.get(oldPni);
        if (oldPrimary === device) {
            this.devicesByUUID.delete(oldPni);
            this.devicesByUUID.set(options.pni, device);
        }
    }
    //
    // Auth
    //
    async auth(username, password) {
        const entry = this.devicesByAuth.get(username);
        if (!entry) {
            debug('auth failed, username=%j is unknown', username);
            return;
        }
        if (entry.password !== password) {
            debug('auth failed, invalid login/password %j:%j', username, password);
        }
        return entry.device;
    }
    //
    // CDN
    //
    async storeAttachment(attachment) {
        const id = constants_1.ATTACHMENT_PREFIX +
            crypto_1.default.createHash('sha256').update(attachment).digest('hex');
        this.attachments.set(id, attachment);
        return id;
    }
    async fetchAttachment(id) {
        return this.attachments.get(id);
    }
    async fetchStickerPack(packId) {
        return this.stickerPacks.get(packId)?.manifest;
    }
    async fetchSticker(packId, stickerId) {
        return this.stickerPacks.get(packId)?.stickers[stickerId];
    }
    async storeStickerPack(pack) {
        this.stickerPacks.set(pack.id.toString('hex'), pack);
    }
    //
    // Messages
    //
    async prepareMultiDeviceMessage(source, targetUUID, messages) {
        if (this.isUnregistered(targetUUID)) {
            return { status: 'unknown' };
        }
        const devices = await this.getAllDevicesByUUID(targetUUID);
        if (devices.length === 0) {
            return { status: 'unknown' };
        }
        const deviceById = new Map();
        for (const device of devices) {
            deviceById.set(device.deviceId, device);
        }
        const result = new Array();
        const extraDevices = new Set();
        const staleDevices = new Set();
        for (const message of messages) {
            const { destinationDeviceId, destinationRegistrationId, } = message;
            const target = deviceById.get(destinationDeviceId);
            if (!target) {
                extraDevices.add(destinationDeviceId);
                continue;
            }
            const uuidKind = target.getUUIDKind(targetUUID);
            deviceById.delete(destinationDeviceId);
            if (target.getRegistrationId(uuidKind) !== destinationRegistrationId) {
                staleDevices.add(destinationDeviceId);
                continue;
            }
            result.push([target, message]);
        }
        if (source && source.uuid === targetUUID) {
            deviceById.delete(source.deviceId);
        }
        if (staleDevices.size !== 0) {
            return { status: 'stale', staleDevices: Array.from(staleDevices) };
        }
        if (extraDevices.size !== 0 || deviceById.size !== 0) {
            return {
                status: 'incomplete',
                missingDevices: Array.from(deviceById.keys()),
                extraDevices: Array.from(extraDevices),
            };
        }
        return { status: 'ok', targetUUID, result };
    }
    async handlePreparedMultiDeviceMessage(source, targetUUID, prepared) {
        for (const [target, message] of prepared) {
            let envelopeType;
            if (message.type === compiled_1.signalservice.Envelope.Type.CIPHERTEXT) {
                envelopeType = EnvelopeType.CipherText;
            }
            else if (message.type === compiled_1.signalservice.Envelope.Type.PREKEY_BUNDLE) {
                envelopeType = EnvelopeType.PreKey;
            }
            else if (message.type === compiled_1.signalservice.Envelope.Type.UNIDENTIFIED_SENDER) {
                envelopeType = EnvelopeType.SealedSender;
            }
            else {
                throw new Error(`Unsupported envelope type: ${message.type}`);
            }
            const uuidKind = target.getUUIDKind(targetUUID);
            await this.handleMessage(source, uuidKind, envelopeType, target, Buffer.from(message.content, 'base64'));
        }
    }
    async addWebSocket(device, socket) {
        debug('adding websocket for device=%s', device.debugId);
        let sockets = this.webSockets.get(device);
        if (!sockets) {
            sockets = new Set();
            this.webSockets.set(device, sockets);
        }
        sockets.add(socket);
        await this.sendQueue(device, socket);
    }
    removeWebSocket(device, socket) {
        debug('removing websocket for device=%s', device.debugId);
        const sockets = this.webSockets.get(device);
        if (!sockets) {
            return;
        }
        sockets.delete(socket);
        if (sockets.size === 0) {
            this.webSockets.delete(device);
        }
    }
    // TODO(indutny): timeout
    async send(target, message) {
        const sockets = this.webSockets.get(target);
        if (sockets) {
            debug('sending message to %d sockets of %s', sockets.size, target.debugId);
            let success = false;
            await Promise.all(Array.from(sockets).map(async (socket) => {
                try {
                    await socket.sendMessage(message);
                    success = true;
                }
                catch (error) {
                    (0, assert_1.default)(error instanceof Error);
                    debug('failed to send message to socket of %s, error %s', target.debugId, error.message);
                }
            }));
            // At least one send should succeed, if not - queue
            if (success) {
                return;
            }
            debug('message couldn\'t be sent to %s', sockets.size, target.debugId);
        }
        debug('queueing message for device=%s', target.debugId);
        await new Promise((resolve, reject) => {
            // NOTE: set and push have to happen in the same tick, otherwise a race
            // condition is possible in `removeWebSocket`.
            let queue = this.messageQueue.get(target);
            if (!queue) {
                queue = [];
                this.messageQueue.set(target, queue);
            }
            queue.push({
                message,
                resolve,
                reject,
            });
        });
        debug('queued message sent to device=%s', target.debugId);
    }
    //
    // Groups
    //
    async createGroup(group) {
        const result = new group_1.ServerGroup({
            zkSecret: this.zkSecret,
            profileOps: new zkgroup_1.ServerZkProfileOperations(this.zkSecret),
            state: group,
        });
        const key = result.publicParams.serialize().toString('base64');
        if (this.groups.get(key)) {
            throw new Error('Duplicate group');
        }
        this.groups.set(key, result);
        return result;
    }
    async modifyGroup({ group, actions, aciCiphertext, pniCiphertext, }) {
        return group.modify(new zkgroup_1.UuidCiphertext(Buffer.from(aciCiphertext)), new zkgroup_1.UuidCiphertext(Buffer.from(pniCiphertext)), actions);
    }
    async getGroup(publicParams) {
        return this.groups.get(publicParams.toString('base64'));
    }
    //
    // Storage
    //
    async getStorageAuth(device) {
        let auth = this.storageAuthByDevice.get(device);
        if (!auth) {
            do {
                auth = {
                    username: crypto_1.default.randomBytes(8).toString('hex'),
                    password: crypto_1.default.randomBytes(8).toString('hex'),
                    device,
                };
            } while (this.storageAuthByUsername.has(auth.username));
            this.storageAuthByDevice.set(device, auth);
            this.storageAuthByUsername.set(auth.username, auth);
            debug('register new storage username=%j', auth.username);
        }
        return {
            username: auth.username,
            password: auth.password,
        };
    }
    async storageAuth(username, password) {
        const auth = this.storageAuthByUsername.get(username);
        if (!auth) {
            debug('auth failed, username=%j is unknown', username);
            return;
        }
        if (auth.password !== password) {
            debug('auth failed, invalid login/password %j:%j', username, password);
        }
        return auth.device;
    }
    async getStorageManifest(device) {
        return this.storageManifestByUuid.get(device.uuid);
    }
    async applyStorageWrite(device, { manifest, clearAll, insertItem, deleteKey, }, shouldNotify = true) {
        if (!manifest) {
            return { error: 'missing `writeOperation.manifest`' };
        }
        if (!manifest.version) {
            return {
                error: 'not updating storage manifest, ' +
                    'missing `writeOperation.manifest.version`',
            };
        }
        const existing = await this.getStorageManifest(device);
        if (existing) {
            // Atomicity
            (0, assert_1.default)(existing.version, 'consistency check');
            if (!manifest.version.eq(existing.version.add(1))) {
                debug('not updating storage manifest, current version=%j new version=%j', existing.version.toNumber(), manifest.version.toNumber());
                return { updated: false, manifest: existing };
            }
        }
        if (clearAll) {
            debug('clearing storage items for=%j', device.debugId);
            await this.clearStorageItems(device);
        }
        const inserts = (insertItem || []).map(async (item) => {
            (0, assert_1.default)(item.key instanceof Uint8Array, 'insertItem.key must be a Buffer');
            (0, assert_1.default)(item.value instanceof Uint8Array, 'insertItem.value must be a Buffer');
            return this.setStorageItem(device, Buffer.from(item.key), Buffer.from(item.value));
        });
        await Promise.all(inserts);
        const deletes = (deleteKey || []).map(async (key) => {
            return this.deleteStorageItem(device, Buffer.from(key));
        });
        await Promise.all(deletes);
        debug('updating storage manifest to version=%j for=%j', manifest.version.toNumber(), device.debugId);
        this.storageManifestByUuid.set(device.uuid, manifest);
        if (shouldNotify) {
            await this.onStorageManifestUpdate(device, manifest.version);
        }
        return { updated: true };
    }
    async clearStorageItems(device) {
        this.storageItemsByUuid.get(device.uuid)?.clear();
    }
    async setStorageItem(device, key, value) {
        let map = this.storageItemsByUuid.get(device.uuid);
        if (!map) {
            map = new Map();
            this.storageItemsByUuid.set(device.uuid, map);
        }
        map.set(key.toString('hex'), value);
    }
    async getStorageItem(device, key) {
        const map = this.storageItemsByUuid.get(device.uuid);
        if (!map) {
            return undefined;
        }
        return map.get(key.toString('hex'));
    }
    async getStorageItems(device, keys) {
        const result = new Array();
        await Promise.all(keys.map(async (key) => {
            const value = await this.getStorageItem(device, key);
            if (value !== undefined) {
                result.push({ key, value });
            }
        }));
        return result;
    }
    async deleteStorageItem(device, key) {
        const map = this.storageItemsByUuid.get(device.uuid);
        if (!map) {
            return;
        }
        map.delete(key.toString('hex'));
    }
    //
    // Utils
    //
    async getDevice(number, deviceId) {
        const list = this.devices.get(number);
        if (!list) {
            return;
        }
        if (deviceId < 1 || deviceId > list.length) {
            return;
        }
        return list[deviceId - 1];
    }
    async getDeviceByUUID(uuid, deviceId) {
        const primary = this.devicesByUUID.get(uuid);
        if (deviceId === undefined || !primary || primary.deviceId === deviceId) {
            return primary;
        }
        if (primary.deviceId !== constants_1.PRIMARY_DEVICE_ID) {
            return undefined;
        }
        return await this.getDevice(primary.number, deviceId);
    }
    async getAllDevicesByUUID(uuid) {
        const primary = this.devicesByUUID.get(uuid);
        if (!primary) {
            return [];
        }
        return this.devices.get(primary.number) || [];
    }
    async getSenderCertificate(device) {
        return (0, crypto_2.generateSenderCertificate)(this.certificate, {
            number: device.number,
            uuid: device.uuid,
            deviceId: device.deviceId,
            identityKey: await device.getIdentityKey(types_1.UUIDKind.ACI),
        });
    }
    async getGroupCredentials({ uuid, pni }, { from, to }) {
        const today = (0, util_1.getTodayInSeconds)();
        if (from > to ||
            from < today ||
            to > today + constants_1.DAY_IN_SECONDS * constants_1.MAX_GROUP_CREDENTIALS_DAYS) {
            throw new Error('Invalid redemption range');
        }
        const auth = new zkgroup_1.ServerZkAuthOperations(this.zkSecret);
        const result = [];
        for (let redemptionTime = from; redemptionTime <= to; redemptionTime += constants_1.DAY_IN_SECONDS) {
            result.push({
                credential: auth.issueAuthCredentialWithPni(uuid, pni, redemptionTime)
                    .serialize().toString('base64'),
                redemptionTime,
            });
        }
        return result;
    }
    async verifyGroupCredentials(publicParams, credential) {
        const auth = new zkgroup_1.ServerZkAuthOperations(this.zkSecret);
        const groupParams = new zkgroup_1.GroupPublicParams(publicParams);
        const presentation = new zkgroup_1.AuthCredentialPresentation(credential);
        auth.verifyAuthCredentialPresentation(groupParams, presentation);
        // TODO(indutny): verify credential timestamp
        return presentation;
    }
    async issueExpiringProfileKeyCredential({ uuid, profileKeyCommitment }, request) {
        if (!profileKeyCommitment) {
            return undefined;
        }
        const today = (0, util_1.getTodayInSeconds)();
        const profile = new zkgroup_1.ServerZkProfileOperations(this.zkSecret);
        return profile.issueExpiringProfileKeyCredential(request, uuid, profileKeyCommitment, today + constants_1.PROFILE_KEY_CREDENTIAL_EXPIRATION).serialize();
    }
    //
    // Private
    //
    set certificate(value) {
        if (this.privCertificate) {
            throw new Error('Certificate already set');
        }
        this.privCertificate = value;
    }
    get certificate() {
        if (!this.privCertificate) {
            throw new Error('Certificate not set');
        }
        return this.privCertificate;
    }
    set zkSecret(value) {
        if (this.privZKSecret) {
            throw new Error('zkgroup secret already set');
        }
        this.privZKSecret = value;
    }
    get zkSecret() {
        if (!this.privZKSecret) {
            throw new Error('zkgroup secret not set');
        }
        return this.privZKSecret;
    }
    async sendQueue(device, socket) {
        let queue = this.messageQueue.get(device);
        if (queue) {
            this.messageQueue.delete(device);
        }
        else {
            queue = [];
        }
        debug('sending queued %d messages to %s', queue.length, device.debugId);
        await Promise.all(queue.map(async (entry) => {
            const { message, resolve, reject } = entry;
            try {
                await socket.sendMessage(message);
            }
            catch (error) {
                (0, assert_1.default)(error instanceof Error);
                reject(error);
                return;
            }
            resolve();
        }));
        debug('queue for %s is empty', device.debugId);
        await socket.sendMessage('empty');
    }
}
exports.Server = Server;
