"use strict";
// Copyright 2022 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createHandler = void 0;
const assert_1 = __importDefault(require("assert"));
const long_1 = __importDefault(require("long"));
const buffer_1 = require("buffer");
const micro_1 = require("micro");
const microrouter_1 = require("microrouter");
const libsignal_client_1 = require("@signalapp/libsignal-client");
const debug_1 = __importDefault(require("debug"));
const util_1 = require("../util");
const types_1 = require("../types");
const compiled_1 = require("../../protos/compiled");
const debug = (0, debug_1.default)('mock:http');
const parsePassword = (req) => {
    return (0, util_1.parseAuthHeader)(req.headers.authorization);
};
const sendDevicesKeys = async (res, uuidKind, devices) => {
    const [primary] = devices;
    (0, assert_1.default)(primary !== undefined, 'Empty device list');
    const identityKey = await primary.getIdentityKey(uuidKind);
    (0, micro_1.send)(res, 200, {
        identityKey: identityKey.serialize().toString('base64'),
        devices: await Promise.all(devices.map(async (device) => {
            const { signedPreKey, preKey } = await device.popSingleUseKey(uuidKind);
            return {
                deviceId: device.deviceId,
                registrationId: device.getRegistrationId(uuidKind),
                signedPreKey: {
                    keyId: signedPreKey.keyId,
                    publicKey: signedPreKey.publicKey.serialize().toString('base64'),
                    signature: signedPreKey.signature.toString('base64'),
                },
                preKey: preKey && {
                    keyId: preKey.keyId,
                    publicKey: preKey.publicKey.serialize().toString('base64'),
                },
            };
        })),
    });
};
const createHandler = (server) => {
    //
    // Unauthorized requests
    //
    const provisionDevice = (0, microrouter_1.put)('/v1/devices/:code', async (req, res) => {
        const { error, username, password } = parsePassword(req);
        if (error) {
            return (0, micro_1.send)(res, 400, { error });
        }
        if (!username || !password) {
            return (0, micro_1.send)(res, 400, { error: 'Invalid authorization header' });
        }
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const body = await (0, micro_1.json)(req);
        if (typeof body.registrationId !== 'number') {
            return (0, micro_1.send)(res, 400, { error: 'Invalid registration id' });
        }
        if (typeof body.pniRegistrationId !== 'number') {
            return (0, micro_1.send)(res, 400, { error: 'Invalid PNI registration id' });
        }
        const device = await server.provisionDevice({
            number: username,
            password,
            provisioningCode: req.params.code,
            registrationId: body.registrationId,
            pniRegistrationId: body.pniRegistrationId,
        });
        return { deviceId: device.deviceId, uuid: device.uuid, pni: device.pni };
    });
    // TODO(indutny): add a route for /v2/keys/:uuid
    const getDeviceKeys = (0, microrouter_1.get)('/v2/keys/:uuid/:deviceId', async (req, res) => {
        const uuid = req.params.uuid;
        const deviceId = parseInt(req.params.deviceId || '', 10);
        if (!uuid || deviceId.toString() !== req.params.deviceId) {
            return (0, micro_1.send)(res, 400, { error: 'Invalid request parameters' });
        }
        const device = await server.getDeviceByUUID(uuid, deviceId);
        if (!device) {
            return (0, micro_1.send)(res, 404, { error: 'Device not found' });
        }
        const uuidKind = device.getUUIDKind(uuid);
        return await sendDevicesKeys(res, uuidKind, [device]);
    });
    const getAllDeviceKeys = (0, microrouter_1.get)('/v2/keys/:uuid(/\\*)', async (req, res) => {
        const uuid = req.params.uuid;
        if (!uuid) {
            return (0, micro_1.send)(res, 400, { error: 'Invalid request parameters' });
        }
        const devices = await server.getAllDevicesByUUID(uuid);
        if (devices.length === 0) {
            return (0, micro_1.send)(res, 404, { error: 'Account not found' });
        }
        const uuidKind = devices[0].getUUIDKind(uuid);
        return await sendDevicesKeys(res, uuidKind, devices);
    });
    //
    // CDN
    //
    const getAttachment = (0, microrouter_1.get)('/attachments/:key/:subkey', async (req, res) => {
        const { key, subkey } = req.params;
        const result = await server.fetchAttachment(`${key}/${subkey}`);
        if (!result) {
            return (0, micro_1.send)(res, 404, { error: 'Attachment not found' });
        }
        return result;
    });
    const getStickerPack = (0, microrouter_1.get)('/stickers/:pack/manifest.proto', async (req, res) => {
        const { pack } = req.params;
        const result = await server.fetchStickerPack(pack);
        if (!result) {
            return (0, micro_1.send)(res, 404, { error: 'Sticker pack not found' });
        }
        return result;
    });
    const getSticker = (0, microrouter_1.get)('/stickers/:pack/full/:sticker', async (req, res) => {
        const { pack, sticker } = req.params;
        const result = await server.fetchSticker(pack, parseInt(sticker, 10));
        if (!result) {
            return (0, micro_1.send)(res, 404, { error: 'Sticker not found' });
        }
        return result;
    });
    const notFound = async (req, res) => {
        debug('Unsupported request %s %s', req.method, req.url);
        return (0, micro_1.send)(res, 404, { error: 'Not supported yet' });
    };
    //
    // Authorized requests
    //
    async function auth(req, res) {
        const { username, password, error } = parsePassword(req);
        if (error) {
            debug('%s %s auth failed, error %j', req.method, req.url, error);
            (0, micro_1.send)(res, 401, { error });
            return;
        }
        const device = await server.auth(username ?? '', password ?? '');
        if (!device) {
            debug('%s %s auth failed, need re-provisioning', req.method, req.url);
            (0, micro_1.send)(res, 401, { error: 'Need re-provisioning' });
            return;
        }
        return device;
    }
    async function groupAuth(req, res) {
        const { error, username, password } = parsePassword(req);
        if (error) {
            (0, micro_1.send)(res, 400, { error });
            return undefined;
        }
        if (!username || !password) {
            (0, micro_1.send)(res, 400, { error: 'Invalid authorization header' });
            return undefined;
        }
        const publicParams = buffer_1.Buffer.from(username, 'hex');
        const credential = buffer_1.Buffer.from(password, 'hex');
        let aciCiphertext;
        let pniCiphertext;
        try {
            const auth = await server.verifyGroupCredentials(publicParams, credential);
            aciCiphertext = auth.getUuidCiphertext();
            const maybePni = auth.getPniCiphertext();
            (0, assert_1.default)(maybePni, 'Auth credentials must have PNI');
            pniCiphertext = maybePni;
        }
        catch (_) {
            (0, micro_1.send)(res, 403, { error: 'Invalid credentials' });
            return undefined;
        }
        return { publicParams, aciCiphertext, pniCiphertext };
    }
    async function groupAuthAndFetch(req, res) {
        const auth = await groupAuth(req, res);
        if (!auth) {
            return;
        }
        const group = await server.getGroup(auth.publicParams);
        if (!group) {
            (0, micro_1.send)(res, 404, { error: 'Group not found' });
            return undefined;
        }
        return { group, ...auth };
    }
    async function storageAuth(req, res) {
        const { error, username, password } = parsePassword(req);
        if (error) {
            (0, micro_1.send)(res, 400, { error });
            return undefined;
        }
        if (!username || !password) {
            (0, micro_1.send)(res, 400, { error: 'Invalid authorization header' });
            return undefined;
        }
        const device = await server.storageAuth(username, password);
        if (!device) {
            debug('%s %s storage auth failed', req.method, req.url);
            (0, micro_1.send)(res, 403, { error: 'Invalid authorization' });
            return undefined;
        }
        return device;
    }
    function uuidKindFromQuery(req) {
        if (req.query.identity === 'pni') {
            return types_1.UUIDKind.PNI;
        }
        return types_1.UUIDKind.ACI;
    }
    const putKeys = (0, microrouter_1.put)('/v2/keys', async (req, res) => {
        const device = await auth(req, res);
        if (!device) {
            return;
        }
        const uuidKind = uuidKindFromQuery(req);
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const body = await (0, micro_1.json)(req);
        try {
            const parseKey = (base64) => {
                return libsignal_client_1.PublicKey.deserialize(buffer_1.Buffer.from(base64, 'base64'));
            };
            await server.updateDeviceKeys(device, uuidKind, {
                identityKey: parseKey(body.identityKey),
                signedPreKey: {
                    keyId: body.signedPreKey.keyId,
                    publicKey: parseKey(body.signedPreKey.publicKey),
                    signature: buffer_1.Buffer.from(body.signedPreKey.signature, 'base64'),
                },
                preKeys: body.preKeys.map((preKey) => {
                    return {
                        keyId: preKey.keyId,
                        publicKey: parseKey(preKey.publicKey),
                    };
                }),
            });
        }
        catch (error) {
            (0, assert_1.default)(error instanceof Error);
            debug('updateDeviceKeys error', error.stack);
            return (0, micro_1.send)(res, 400, { error: error.message });
        }
        return { ok: true };
    });
    const getKeys = (0, microrouter_1.get)('/v2/keys', async (req, res) => {
        const device = await auth(req, res);
        if (!device) {
            return;
        }
        const uuidKind = uuidKindFromQuery(req);
        return { count: await device.getSingleUseKeyCount(uuidKind) };
    });
    const whoami = (0, microrouter_1.get)('/v1/accounts/whoami', async (req, res) => {
        const device = await auth(req, res);
        if (!device) {
            return;
        }
        return { uuid: device.uuid, pni: device.pni, number: device.number };
    });
    const putChallenge = (0, microrouter_1.put)('/v1/challenge', async (req, res) => {
        const device = await auth(req, res);
        if (!device) {
            return;
        }
        return { ok: true };
    });
    //
    // GV2
    //
    const getGroup = (0, microrouter_1.get)('/v1/groups', async (req, res) => {
        const auth = await groupAuthAndFetch(req, res);
        if (!auth) {
            return;
        }
        const { group } = auth;
        return (0, micro_1.send)(res, 200, compiled_1.signalservice.Group.encode(group.state).finish());
    });
    const getGroupVersion = (0, microrouter_1.get)('/v1/groups/joined_at_version', async (req, res) => {
        const auth = await groupAuthAndFetch(req, res);
        if (!auth) {
            return;
        }
        const { group, aciCiphertext } = auth;
        const member = group.getMember(aciCiphertext);
        if (!member) {
            return (0, micro_1.send)(res, 403, { error: 'Not a member of this group' });
        }
        return (0, micro_1.send)(res, 200, compiled_1.signalservice.Member.encode({
            joinedAtVersion: member.joinedAtVersion,
        }).finish());
    });
    const getGroupLogs = (0, microrouter_1.get)('/v1/groups/logs/:since', async (req, res) => {
        const auth = await groupAuthAndFetch(req, res);
        if (!auth) {
            return;
        }
        const { group, aciCiphertext } = auth;
        const member = group.getMember(aciCiphertext);
        if (!member) {
            return (0, micro_1.send)(res, 403, { error: 'Not a member of this group' });
        }
        const since = parseInt(req.params.since, 10);
        if (since < (member.joinedAtVersion ?? 0)) {
            return (0, micro_1.send)(res, 403, { error: '`since` is before joinedAtVersion' });
        }
        return (0, micro_1.send)(res, 200, compiled_1.signalservice.GroupChanges.encode(group.getChangesSince(since)).finish());
    });
    // TODO(indutny): implement me
    const createGroup = (0, microrouter_1.put)('/v1/groups', async (req, res) => {
        const auth = await groupAuth(req, res);
        if (!auth) {
            return;
        }
        const groupData = compiled_1.signalservice.Group.decode(buffer_1.Buffer.from(await (0, micro_1.buffer)(req)));
        if (!groupData.title) {
            return (0, micro_1.send)(res, 400, { error: 'Missing group title' });
        }
        if (!groupData.publicKey || !auth.publicParams.equals(groupData.publicKey)) {
            return (0, micro_1.send)(res, 400, { error: 'Invalid group public key' });
        }
        await server.createGroup(groupData);
        // TODO(indutny): verify that creator is a member
        return (0, micro_1.send)(res, 200);
    });
    const modifyGroup = (0, microrouter_1.patch)('/v1/groups', async (req, res) => {
        const auth = await groupAuthAndFetch(req, res);
        if (!auth) {
            return;
        }
        const actions = compiled_1.signalservice.GroupChange.Actions.decode(buffer_1.Buffer.from(await (0, micro_1.buffer)(req)));
        const { group, aciCiphertext, pniCiphertext } = auth;
        try {
            const modifyResult = await server.modifyGroup({
                group,
                aciCiphertext: aciCiphertext.serialize(),
                pniCiphertext: pniCiphertext.serialize(),
                actions,
            });
            if (modifyResult.conflict) {
                return (0, micro_1.send)(res, 409, { error: 'Conflict' });
            }
            return (0, micro_1.send)(res, 200, compiled_1.signalservice.GroupChange.encode(modifyResult.signedChange).finish());
        }
        catch (error) {
            (0, assert_1.default)(error instanceof Error);
            debug('Failed to modify group', error.stack);
            // TODO(indutny): would be nice to give 403 here
            return (0, micro_1.send)(res, 500, { error: error.stack });
        }
        return (0, micro_1.send)(res, 200, compiled_1.signalservice.Group.encode(group.state).finish());
    });
    //
    // Storage Service
    //
    const getStorageManifest = (0, microrouter_1.get)('/v1/storage/manifest', async (req, res) => {
        const device = await storageAuth(req, res);
        if (!device) {
            return;
        }
        const manifest = await server.getStorageManifest(device);
        if (!manifest) {
            return (0, micro_1.send)(res, 404, { error: 'Manifest not found' });
        }
        return (0, micro_1.send)(res, 200, compiled_1.signalservice.StorageManifest.encode(manifest).finish());
    });
    const getStorageManifestByVersion = (0, microrouter_1.get)('/v1/storage/manifest/version/:after', async (req, res) => {
        const device = await storageAuth(req, res);
        if (!device) {
            return;
        }
        const after = long_1.default.fromString(req.params.after);
        const manifest = await server.getStorageManifest(device);
        if (!manifest?.version?.gt(after)) {
            return (0, micro_1.send)(res, 204);
        }
        return (0, micro_1.send)(res, 200, compiled_1.signalservice.StorageManifest.encode(manifest).finish());
    });
    const putStorage = (0, microrouter_1.put)('/v1/storage/', async (req, res) => {
        const device = await storageAuth(req, res);
        if (!device) {
            return;
        }
        const writeOperation = compiled_1.signalservice.WriteOperation.decode(buffer_1.Buffer.from(await (0, micro_1.buffer)(req)));
        const result = await server.applyStorageWrite(device, writeOperation);
        if ('error' in result) {
            return (0, micro_1.send)(res, 400, { error: result.error });
        }
        if (!result.updated) {
            return (0, micro_1.send)(res, 409, compiled_1.signalservice.StorageManifest.encode(result.manifest).finish());
        }
        return (0, micro_1.send)(res, 200);
    });
    const putStorageRead = (0, microrouter_1.put)('/v1/storage/read', async (req, res) => {
        const device = await storageAuth(req, res);
        if (!device) {
            return;
        }
        const readOperation = compiled_1.signalservice.ReadOperation.decode(buffer_1.Buffer.from(await (0, micro_1.buffer)(req)));
        const keys = (readOperation.readKey || []).map((key) => buffer_1.Buffer.from(key));
        const items = await server.getStorageItems(device, keys);
        if (!items) {
            return (0, micro_1.send)(res, 413, { error: 'Requested too many items' });
        }
        return (0, micro_1.send)(res, 200, compiled_1.signalservice.StorageItems.encode({
            items,
        }).finish());
    });
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    const dummyAuth = (response) => {
        return async (req, res) => {
            const device = await auth(req, res);
            if (!device) {
                return;
            }
            return response;
        };
    };
    const notFoundAfterAuth = async (req, res) => {
        const device = await auth(req, res);
        if (!device) {
            return;
        }
        debug('Unsupported request %s %s', req.method, req.url);
        return (0, micro_1.send)(res, 404, { error: 'Not supported yet' });
    };
    const routes = (0, microrouter_1.router)(
    // Sure, why not
    (0, microrouter_1.get)('/v1/config', dummyAuth({ config: [] })), (0, microrouter_1.put)('/v1/devices/unauthenticated_delivery', dummyAuth({ ok: true })), (0, microrouter_1.put)('/v1/devices/capabilities', dummyAuth({ ok: true })), 
    // TODO(indutny): support nameless devices? They use different route
    provisionDevice, getDeviceKeys, getAllDeviceKeys, getAttachment, getStickerPack, getSticker, putKeys, getKeys, whoami, putChallenge, 
    // Technically these should live on a separate server, but who cares
    getGroup, getGroupVersion, getGroupLogs, createGroup, modifyGroup, getStorageManifest, getStorageManifestByVersion, putStorage, putStorageRead, 
    // TODO(indutny): support this
    (0, microrouter_1.get)('/v1/groups/token', notFound), (0, microrouter_1.get)('/stickers/', notFound), (0, microrouter_1.get)('/*', notFoundAfterAuth), (0, microrouter_1.put)('/*', notFoundAfterAuth));
    return (req, res) => {
        debug('got request %s %s', req.method, req.url);
        return routes(req, res);
    };
};
exports.createHandler = createHandler;
