"use strict";
// Copyright 2022 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Connection = void 0;
const assert_1 = __importDefault(require("assert"));
const buffer_1 = require("buffer");
const url_1 = require("url");
const crypto_1 = require("crypto");
const debug_1 = __importDefault(require("debug"));
const zkgroup_1 = require("@signalapp/libsignal-client/zkgroup");
const compiled_1 = require("../../../protos/compiled");
const crypto_2 = require("../../crypto");
const util_1 = require("../../util");
const service_1 = require("./service");
const router_1 = require("./router");
const debug = (0, debug_1.default)('mock:ws:connection');
class Connection extends service_1.Service {
    constructor(request, ws, server) {
        super(ws);
        this.request = request;
        this.server = server;
        this.router = new router_1.Router();
        const getProfile = async (params, _, headers, { credentialType } = {}) => {
            const uuid = params.uuid;
            const target = await this.server.getDeviceByUUID(uuid);
            if (!target) {
                return [404, { error: 'Device not found' }];
            }
            if (this.server.isUnregistered(uuid)) {
                return [404, { error: 'Unregistered' }];
            }
            const accessError = this.checkAccessKey(target, headers);
            if (accessError !== undefined) {
                return [401, { error: accessError }];
            }
            let credential;
            if (params.request) {
                const request = new zkgroup_1.ProfileKeyCredentialRequest(buffer_1.Buffer.from(params.request, 'hex'));
                if (credentialType === 'expiringProfileKey') {
                    credential = await this.server.issueExpiringProfileKeyCredential(target, request);
                }
                else {
                    return [400, { error: 'Unsupported credential type' }];
                }
            }
            const uuidKind = target.getUUIDKind(uuid);
            const identityKey = await target.getIdentityKey(uuidKind);
            return [200, {
                    name: target.profileName,
                    identityKey: identityKey.serialize().toString('base64'),
                    unrestrictedUnidentifiedAccess: false,
                    unidentifiedAccess: target.accessKey ?
                        (0, crypto_2.generateAccessKeyVerifier)(target.accessKey) : undefined,
                    capabilities: {
                        announcementGroup: true,
                        'gv2-3': true,
                        'gv1-migration': true,
                        senderKey: true,
                    },
                    credential: credential?.toString('base64'),
                }];
        };
        this.router.get('/v1/profile/:uuid', getProfile);
        this.router.get('/v1/profile/:uuid/:version', getProfile);
        this.router.get('/v1/profile/:uuid/:version/:request', getProfile);
        const requireAuth = (handler) => {
            return async (params, body, headers) => {
                if (!this.device) {
                    return [401, { error: 'Not authorized' }];
                }
                return handler(params, body, headers);
            };
        };
        this.router.get('/v1/config', requireAuth(async () => {
            return [200, {
                    config: [
                        { name: 'desktop.gv2', enabled: true },
                        { name: 'desktop.gv2Admin', enabled: true },
                        { name: 'desktop.internalUser', enabled: true },
                        { name: 'desktop.sendSenderKey2', enabled: true },
                        { name: 'desktop.sendSenderKey3', enabled: true },
                        { name: 'desktop.senderKey.retry', enabled: true },
                        { name: 'desktop.senderKey.send', enabled: true },
                        { name: 'desktop.storage', enabled: true },
                        { name: 'desktop.storageWrite3', enabled: true },
                        { name: 'desktop.messageRequests', enabled: true },
                        { name: 'desktop.pnp', enabled: true },
                        {
                            name: 'global.groupsv2.maxGroupSize',
                            value: '32',
                            enabled: true,
                        },
                        {
                            name: 'global.groupsv2.groupSizeHardLimit',
                            value: '64',
                            enabled: true,
                        },
                    ],
                }];
        }));
        this.router.put('/v1/messages/multi_recipient', async (_params, body) => {
            if (!body) {
                return [400, { error: 'Missing body' }];
            }
            const { recipients, commonMaterial, } = (0, util_1.parseMultiRecipientMessage)(buffer_1.Buffer.from(body));
            const listByUUID = new Map();
            for (const recipient of recipients) {
                const { uuid, deviceId, registrationId, material, } = recipient;
                let list = listByUUID.get(uuid);
                if (!list) {
                    list = [];
                    listByUUID.set(uuid, list);
                }
                list.push({
                    type: compiled_1.signalservice.Envelope.Type.UNIDENTIFIED_SENDER,
                    destinationDeviceId: deviceId,
                    destinationRegistrationId: registrationId,
                    content: (0, util_1.combineMultiRecipientMessage)({
                        material,
                        commonMaterial,
                    }).toString('base64'),
                });
            }
            // TODO(indutny): verify access key xor
            const results = await Promise.all(Array.from(listByUUID.entries()).map(async ([uuid, messages]) => {
                return {
                    uuid,
                    prepared: await this.server.prepareMultiDeviceMessage(undefined, uuid, messages),
                };
            }));
            const incomplete = results.filter(({ prepared }) => prepared.status === 'incomplete');
            if (incomplete.length !== 0) {
                return [
                    409,
                    incomplete.map(({ uuid, prepared }) => {
                        assert_1.default.ok(prepared.status === 'incomplete');
                        return {
                            uuid,
                            devices: {
                                missingDevices: prepared.missingDevices,
                                extraDevices: prepared.extraDevices,
                            },
                        };
                    }),
                ];
            }
            const stale = results.filter(({ prepared }) => prepared.status === 'stale');
            if (stale.length !== 0) {
                return [
                    410,
                    stale.map(({ uuid, prepared }) => {
                        assert_1.default.ok(prepared.status === 'stale');
                        return { uuid, devices: { staleDevices: prepared.staleDevices } };
                    }),
                ];
            }
            const uuids404 = results.filter(({ prepared }) => prepared.status === 'unknown').map(({ uuid }) => uuid);
            const ok = results.filter(({ prepared }) => prepared.status === 'ok');
            await Promise.all(ok.map(({ prepared }) => {
                assert_1.default.ok(prepared.status === 'ok');
                return this.server.handlePreparedMultiDeviceMessage(undefined, prepared.targetUUID, prepared.result);
            }));
            return [200, { uuids404 }];
        });
        this.router.put('/v1/messages/:uuid', async (params, body, headers, query = {}) => {
            if (!body) {
                return [400, { error: 'Missing body' }];
            }
            const { messages } = JSON.parse(buffer_1.Buffer.from(body).toString());
            const targetUUID = params.uuid;
            const target = await this.server.getDeviceByUUID(targetUUID);
            if (!target) {
                return [404, { error: 'Device not found' }];
            }
            if (query.story !== 'true') {
                const accessError = this.checkAccessKey(target, headers);
                if (accessError !== undefined) {
                    return [401, { error: accessError }];
                }
            }
            if (this.server.isUnregistered(targetUUID)) {
                return [404, { error: 'Unregistered' }];
            }
            if (this.device &&
                this.server.isSendRateLimited({
                    source: this.device.uuid,
                    target: targetUUID,
                })) {
                return [428, { token: 'token', options: ['recaptcha'] }];
            }
            const prepared = await this.server.prepareMultiDeviceMessage(this.device, params.uuid, messages);
            switch (prepared.status) {
                case 'ok':
                    await this.server.handlePreparedMultiDeviceMessage(this.device, prepared.targetUUID, prepared.result);
                    return [200, { ok: true }];
                case 'unknown':
                    return [404, { error: 'Not found' }];
                case 'incomplete':
                    return [409, {
                            missingDevices: prepared.missingDevices,
                            extraDevices: prepared.extraDevices,
                        }];
                case 'stale':
                    return [410, { staleDevices: prepared.staleDevices }];
            }
        });
        this.router.put('/v1/devices/capabilities', requireAuth(async () => {
            return [200, { ok: true }];
        }));
        this.router.put('/v1/devices/unauthenticated_delivery', requireAuth(async () => {
            return [200, { ok: true }];
        }));
        this.router.get('/v1/certificate/delivery', requireAuth(async () => {
            const device = this.device;
            if (!device) {
                throw new Error('No support for unauthorized delivery');
            }
            const certificate = await this.server.getSenderCertificate(device);
            return [
                200,
                { certificate: certificate.serialize().toString('base64') },
            ];
        }));
        this.router.put('/v1/devices/:code', async (params, body, headers) => {
            const { error, username, password } = (0, util_1.parseAuthHeader)(headers.authorization);
            if (error) {
                return [400, { error }];
            }
            if (!username || !password) {
                return [400, { error: 'Invalid authorization header' }];
            }
            if (!body) {
                return [400, { error: 'Missing body' }];
            }
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const json = JSON.parse(buffer_1.Buffer.from(body).toString());
            if (typeof json.registrationId !== 'number') {
                return [400, { error: 'Invalid registration id' }];
            }
            const device = await server.provisionDevice({
                number: username,
                password,
                provisioningCode: params.code,
                registrationId: json.registrationId,
                pniRegistrationId: json.pniRegistrationId,
            });
            return [200, {
                    deviceId: device.deviceId,
                    uuid: device.uuid,
                    pni: device.pni,
                }];
        });
        //
        // Groups
        //
        this.router.get('/v1/certificate/auth/group', async (_params, _body, _headers, query = {}) => {
            const device = this.device;
            if (!device) {
                throw new Error('No support for unauthorized delivery');
            }
            const { redemptionStartSeconds: from, redemptionEndSeconds: to, } = query;
            return [
                200,
                {
                    credentials: await this.server.getGroupCredentials(device, {
                        from: parseInt(from, 10),
                        to: parseInt(to, 10),
                    }),
                    pni: device.pni,
                },
            ];
        });
        //
        // Storage Service
        //
        this.router.get('/v1/storage/auth', async () => {
            const device = this.device;
            if (!device) {
                throw new Error('Storage credentials require authorization');
            }
            return [200, await server.getStorageAuth(device)];
        });
    }
    async start() {
        debug('Got a websocket connection', this.request.url);
        const url = this.request.url;
        if (!url) {
            throw new Error('Request must have url');
        }
        if (url.startsWith('/v1/websocket/provisioning')) {
            const uuid = await this.server.generateUUID();
            try {
                await this.handleProvision(uuid);
            }
            catch (error) {
                await this.server.releaseUUID(uuid);
                throw error;
            }
            return;
        }
        if (url.startsWith('/v1/websocket/?')) {
            return await this.handleNormal(url);
        }
    }
    async sendMessage(message) {
        let response;
        if (message === 'empty') {
            response = await this.send('PUT', '/api/v1/queue/empty', {});
        }
        else {
            response = await this.send('PUT', '/api/v1/message', {
                body: message,
            });
        }
        assert_1.default.strictEqual(response.status, 200, `WebSocket send error ${response.status} ${response.message}`);
    }
    //
    // Service implementation
    //
    async handleRequest(request) {
        return this.router.run(request);
    }
    //
    // Private
    //
    async handleProvision(uuid) {
        {
            const { status } = await this.send('PUT', '/v1/address', {
                body: compiled_1.signalservice.ProvisioningUuid.encode({
                    uuid,
                }).finish(),
            });
            assert_1.default.strictEqual(status, 200);
        }
        {
            const { envelope } = await this.server.getProvisioningResponse(uuid);
            const { status } = await this.send('PUT', '/v1/message', {
                body: envelope,
            });
            assert_1.default.strictEqual(status, 200);
        }
    }
    async handleNormal(url) {
        const query = (0, url_1.parse)(url, true).query || {};
        if (!query.login ||
            Array.isArray(query.login) ||
            !query.password ||
            Array.isArray(query.password)) {
            debug('Unauthorized WebSocket connection');
            return;
        }
        const device = await this.server.auth(query.login, query.password);
        if (!device) {
            debug('Invalid WebSocket credentials %j', query);
            this.ws.close();
            return;
        }
        this.device = device;
        this.router.setIsAuthenticated(true);
        this.ws.once('close', () => {
            this.server.removeWebSocket(device, this);
        });
        await this.server.addWebSocket(device, this);
    }
    checkAccessKey(target, headers) {
        if (this.device) {
            // Authenticated
        }
        else if (!target.accessKey || !headers['unidentified-access-key']) {
            return 'Not authenticated';
        }
        else {
            const accessKey = buffer_1.Buffer.from(headers['unidentified-access-key'], 'base64');
            if (!(0, crypto_1.timingSafeEqual)(accessKey, target.accessKey)) {
                return 'Invalid access key';
            }
        }
        return undefined;
    }
}
exports.Connection = Connection;
