"use strict";
// Copyright 2022 Signal Messenger, LLC
// SPDX-License-Identifier: AGPL-3.0-only
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateRegistrationId = exports.getTodayInSeconds = exports.combineMultiRecipientMessage = exports.parseMultiRecipientMessage = exports.addressToString = exports.PromiseQueue = exports.parseAuthHeader = exports.generateRandomE164 = void 0;
const assert_1 = __importDefault(require("assert"));
const protobufjs_1 = require("protobufjs");
const uuid_1 = require("uuid");
const constants_1 = require("./constants");
function generateRandomE164() {
    // Generate random number
    let number = '+141549';
    for (let i = 0; i < 5; i++) {
        number += Math.floor(Math.random() * 10).toString();
    }
    return number;
}
exports.generateRandomE164 = generateRandomE164;
const MULTI_RECIPIENT_MESSAGE_VERSION = 0x22;
const MULTI_RECIPIENT_UUID_LEN = 16;
const MULTI_RECIPIENT_SHARED_MATERIAL_LEN = 48;
function parseAuthHeader(header) {
    if (!header) {
        return { error: 'Missing Authorization header' };
    }
    const [basic, base64] = header.split(/\s+/g, 2);
    if (basic.toLowerCase() !== 'basic') {
        return { error: `Unsupported authorization type ${basic}` };
    }
    let username;
    let password;
    try {
        const decoded = Buffer.from(base64, 'base64').toString();
        [username, password] = decoded.split(':', 2);
    }
    catch (error) {
        (0, assert_1.default)(error instanceof Error);
        return { error: error.message };
    }
    if (!username) {
        return { error: 'Missing username' };
    }
    if (!password) {
        return { error: 'Missing password' };
    }
    return { username, password };
}
exports.parseAuthHeader = parseAuthHeader;
class PromiseQueue {
    constructor(config = {}) {
        this.entries = [];
        this.resolvers = [];
        this.defaultTimeout = config.timeout;
    }
    async pushAndWait(value, timeout = this.defaultTimeout) {
        // We were waiting for `.shift()` already
        const resolveEntry = this.resolvers.shift();
        if (resolveEntry) {
            resolveEntry(value);
            return;
        }
        // Not waiting for `.shift()` - queue.
        return await new Promise((resolve, reject) => {
            let timer;
            const entry = {
                value,
                resolvePush() {
                    if (timer !== undefined) {
                        clearTimeout(timer);
                    }
                    timer = undefined;
                    resolve();
                },
            };
            const cancel = () => {
                const index = this.entries.indexOf(entry);
                if (index === -1) {
                    throw new Error('PromiseQueue entries bookkeeping error');
                }
                this.entries.splice(index, 1);
                reject(new Error('PromiseQueue pushAndWait timeout'));
            };
            if (timeout !== undefined) {
                timer = setTimeout(cancel, timeout);
            }
            this.entries.push(entry);
        });
    }
    push(value) {
        // We were waiting for `.shift()` already
        const resolveEntry = this.resolvers.shift();
        if (resolveEntry) {
            resolveEntry(value);
            return;
        }
        this.entries.push({ value });
    }
    async shift(timeout = this.defaultTimeout) {
        // `.pushAndWait()` was called before us
        const entry = this.entries.shift();
        if (entry) {
            if (entry.resolvePush) {
                entry.resolvePush();
            }
            return entry.value;
        }
        return await new Promise((resolve, reject) => {
            let timer;
            const resolveEntry = (value) => {
                if (timer !== undefined) {
                    clearTimeout(timer);
                }
                timer = undefined;
                resolve(value);
            };
            const cancel = () => {
                const index = this.resolvers.indexOf(resolveEntry);
                if (index === -1) {
                    throw new Error('PromiseQueue resolvers bookkeeping error');
                }
                this.resolvers.splice(index, 1);
                reject(new Error('PromiseQueue shift timeout'));
            };
            if (timeout !== undefined) {
                timer = setTimeout(cancel, timeout);
            }
            this.resolvers.push(resolveEntry);
        });
    }
}
exports.PromiseQueue = PromiseQueue;
function addressToString(address) {
    return `${address.name()}.${address.deviceId()}`;
}
exports.addressToString = addressToString;
function parseMultiRecipientMessage(message) {
    if (message[0] !== MULTI_RECIPIENT_MESSAGE_VERSION) {
        throw new Error('Invalid multi-recipient message');
    }
    const reader = new protobufjs_1.BufferReader(message);
    reader.skip(1);
    const count = reader.uint32();
    const recipients = new Array();
    while (recipients.length < count) {
        const uuid = (0, uuid_1.stringify)(message.slice(reader.pos, reader.pos + MULTI_RECIPIENT_UUID_LEN));
        reader.skip(MULTI_RECIPIENT_UUID_LEN);
        const deviceId = reader.uint32();
        const registrationId = message.readUInt16BE(reader.pos);
        reader.skip(2);
        const material = message.slice(reader.pos, reader.pos + MULTI_RECIPIENT_SHARED_MATERIAL_LEN);
        assert_1.default.strictEqual(material.length, MULTI_RECIPIENT_SHARED_MATERIAL_LEN);
        reader.skip(MULTI_RECIPIENT_SHARED_MATERIAL_LEN);
        recipients.push({ uuid, deviceId, registrationId, material });
    }
    const commonMaterial = message.slice(reader.pos);
    return { recipients, commonMaterial };
}
exports.parseMultiRecipientMessage = parseMultiRecipientMessage;
function combineMultiRecipientMessage({ material, commonMaterial }) {
    return Buffer.concat([
        Buffer.from([MULTI_RECIPIENT_MESSAGE_VERSION]),
        material,
        commonMaterial,
    ]);
}
exports.combineMultiRecipientMessage = combineMultiRecipientMessage;
function getTodayInSeconds() {
    const now = Date.now() / 1000;
    return now - (now % constants_1.DAY_IN_SECONDS);
}
exports.getTodayInSeconds = getTodayInSeconds;
function generateRegistrationId() {
    return Math.max(1, (Math.random() * 0x4000) | 0);
}
exports.generateRegistrationId = generateRegistrationId;
